from time import sleep

from scapy.all import (
    Ether,
    Packet,
    X3BytesField,
    XByteField,
    XIntField,
    XLEShortField,
    XNBytesField,
    XShortField,
    sendp,
)

from pyslac.enums import (
    CM_ATTEN_CHAR,
    CM_ATTEN_PROFILE,
    CM_MNBC_SOUND,
    CM_SET_KEY,
    CM_SLAC_MATCH,
    CM_SLAC_PARM,
    CM_START_ATTEN_CHAR,
    ETH_TYPE_HPAV,
    MMTYPE_CNF,
    MMTYPE_IND,
    MMTYPE_REQ,
    MMTYPE_RSP,
    SLAC_SETTLE_TIME,
)
from pyslac.utils import get_if_hwaddr

BROADCAST_ADDR = "FF:FF:FF:FF:FF:FF"
ATHEROS_CHIP_MAC = "00:b0:52:00:00:01"
IFACE = "enp0s3"


class HomePlugHeader(Packet):
    name = "HomePlugHeader "
    fields_desc = [
        XByteField("mmv", 1),
        XLEShortField("mm_type", CM_SET_KEY | MMTYPE_REQ),
        XByteField("fmsn", 0),
        XByteField("fmid", 0),
    ]


class SetKeyRequest(Packet):
    name = "SetKeyRequest Payload "
    fields_desc = [
        XByteField("key_type", 1),
        XIntField("my_nonce", 0xAAAAAAAA),
        XIntField("your_nonce", 0x00000000),
        XByteField("pid", 4),
        XShortField("prn", 0x0000),
        XByteField("pmn", 0),
        XByteField("cco_cap", 0),
        XIntField("nid_first_4_bytes", 0x026BCBA5),
        X3BytesField("nid_last_3bytes", 0x354E08),
        XByteField("new_eks", 1),
        XNBytesField("new_key", 0xB59319D7E8157BA001B018669CCEE30D, 16),
        X3BytesField("rsvd", 0x000000),
    ]


class SetKeyConfirmation(Packet):
    name = "SetKeyResponse Payload "
    fields_desc = [
        XByteField("result", 0),
        XIntField("my_nonce", 0xAAAAAAAA),
        XIntField("your_nonce", 0x00000000),
        XByteField("pid", 4),
        XShortField("prn", 0x0000),
        XByteField("pmn", 0),
        XByteField("cco_cap", 0),
        XNBytesField("rsvd", 0, 27),
    ]


# EV Messages to Send
class SlacParmReq(Packet):
    name = "Slac Parm Req Payload "
    fields_desc = [
        XByteField("application_type", 0x00),
        XByteField("security_type", 0x00),
        XNBytesField("run_id", 0, 8),
        XNBytesField("rsvd", 0, 31),
    ]


class StartAttenChar(Packet):
    name = "Start Atten Characterization Payload "
    fields_desc = [
        XByteField("application_type", 0x00),
        XByteField("security_type", 0x00),
        XByteField(
            "num_sounds", 2
        ),  # This defines the number of sounds that the EV will send to the EVSE.
        # And overrides the expected sounds defined by enum SLAC_MSOUNDS
        XByteField(
            "time_out", 100
        ),  # This defines the time to 10 secs (100 * 100 ms) that the EV must
        # deliver the mnbc sounds before the EVSE times out
        XByteField("resp_type", 0x01),
        XNBytesField("forwarding_sta", 0, 6),
        XNBytesField("run_id", 0, 8),
        XNBytesField("rsvd", 0, 22),
    ]


class MnbcSound(Packet):
    name = "MnbcSound Payload "
    fields_desc = [
        XByteField("application_type", 0x00),
        XByteField("security_type", 0x00),
        XNBytesField("sender_id", 0, 17),
        XByteField("cnt", 2),
        XNBytesField("run_id", 0, 8),
        XNBytesField("rsvd", 0, 8),
        XNBytesField("rnd", 0, 16),
    ]


class AttenProfile(Packet):
    name = "Atten Profile Payload "
    fields_desc = [
        XNBytesField("pev_mac", 0, 6),
        XByteField("num_groups", 3),
        XByteField("rsvd", 0),
        XByteField("aag1", 40),
        XByteField("aag2", 36),
        XByteField("aag3", 40),
        XNBytesField("rsvd1", 0, 60 - 30),
    ]


class AttenCharResp(Packet):
    name = "Start Atten Characterization Payload "
    fields_desc = [
        XByteField("application_type", 0x00),
        XByteField("security_type", 0x00),
        XNBytesField("source_address", 0, 6),
        XNBytesField("run_id", 0, 8),
        XNBytesField("source_id", 0, 17),
        XNBytesField("resp_id", 0, 17),
        XByteField("result", 0),
    ]


class SlacMatch(Packet):
    name = "Slac Match Request Payload "
    fields_desc = [
        XByteField("application_type", 0x00),
        XByteField("security_type", 0x00),
        XNBytesField("mvf_length", 0x003E, 2),
        XNBytesField("pev_id", 0, 17),
        XNBytesField("pev_mac", 0, 6),
        XNBytesField("evse_id", 0, 17),
        XNBytesField("evse_mac", 0, 6),
        XNBytesField("run_id", 0, 8),
        XNBytesField("rsvd", 0, 8),
    ]


host_mac = get_if_hwaddr(IFACE, to_mac_fmt=True)
host_mac_bytes = get_if_hwaddr(IFACE, to_mac_fmt=False)
host_mac_int = int.from_bytes(host_mac_bytes, "big")
eth_header = Ether(src=host_mac, dst=ATHEROS_CHIP_MAC, type=ETH_TYPE_HPAV)


homeplug_header = HomePlugHeader()
set_key_request = SetKeyRequest()
frame_rsp = eth_header / homeplug_header / set_key_request

# we can use this way of padding or the rsvd field
# padstr = '\x00' * (60 - len(eth_header) - len(homeplug_header) - len(key_request_payload))  # noqa: E501
# pad = Padding(load=padstr) # a bunch of 0-bytes
# frame = eth_header/homeplug_header/set_key_request/pad

# SetKey CNF
homeplug_header.mm_type = CM_SET_KEY | MMTYPE_CNF
set_key_confirmation = SetKeyConfirmation()
frame_rsp = eth_header / homeplug_header / set_key_confirmation
sendp(frame_rsp, iface=IFACE)
# We need to wait SLAC_SETTLE_TIME because is the time the EVSE will wait
# after receiving the SET_KEY_CNF for the HLE to settle. In a real EV
# simulator, this wouldnt be  needed
sleep(SLAC_SETTLE_TIME)
# Slac Parm Req
homeplug_header.mm_type = CM_SLAC_PARM | MMTYPE_REQ
eth_header.dst = BROADCAST_ADDR
slac_parm_req = SlacParmReq()
frame_rsp = eth_header / homeplug_header / slac_parm_req
sleep(10)
sendp(frame_rsp, iface=IFACE)

# Start Atten Char
homeplug_header.mm_type = CM_START_ATTEN_CHAR | MMTYPE_IND
eth_header.dst = BROADCAST_ADDR
start_atten_char = StartAttenChar()
frame_rsp = eth_header / homeplug_header / start_atten_char
# The EV may send 3 start atten char, but the QCA chip will forward only 1
# to the application
sendp(frame_rsp, iface=IFACE)

# MNBC Sound
# Send 2 times, since we defined 2 num of sounds in StartAttenChar
homeplug_header.mm_type = CM_MNBC_SOUND | MMTYPE_IND
eth_header.dst = BROADCAST_ADDR
ev_cm_mnbc_sound = MnbcSound()
frame_rsp_mnbc = eth_header / homeplug_header / ev_cm_mnbc_sound

# AttenProfile
#  Send 2 times, since we defined 2 num of sounds in StartAttenChar
homeplug_header.mm_type = CM_ATTEN_PROFILE | MMTYPE_IND
eth_header.src = ATHEROS_CHIP_MAC
eth_header.dst = host_mac
atten_profile = AttenProfile(pev_mac=host_mac_int)
frame_rsp_atten = eth_header / homeplug_header / atten_profile

sendp(frame_rsp_mnbc, iface=IFACE)
sleep(0.1)
sendp(frame_rsp_atten, iface=IFACE)
sleep(0.1)
sendp(frame_rsp_mnbc, iface=IFACE)
sleep(0.1)
sendp(frame_rsp_atten, iface=IFACE)
sleep(0.1)


# EV Receives a Atten Char Indicator and shall respond with Response
# Unicast
homeplug_header.mm_type = CM_ATTEN_CHAR | MMTYPE_RSP
eth_header.src = host_mac
eth_header.dst = host_mac
atten_char_rsp = AttenCharResp(source_address=host_mac_int)
frame_rsp = eth_header / homeplug_header / atten_char_rsp
sendp(frame_rsp, iface=IFACE)
sleep(0.2)


# EV does the Atten threshold calculation and evaluation, then sends a Match
homeplug_header.mm_type = CM_SLAC_MATCH | MMTYPE_REQ
slac_match = SlacMatch(pev_mac=host_mac_int, evse_mac=host_mac_int)
frame_rsp = eth_header / homeplug_header / slac_match
sendp(frame_rsp, iface=IFACE)
# Then the EV Should receive a MAtch conf with the NMK and NID so he can
# join the network



import asyncio
import logging
import socket
import struct
import time
from fcntl import ioctl
from hashlib import sha256
from math import copysign
from sys import platform
from typing import Any, Awaitable, List

logging.basicConfig(level=logging.DEBUG)
logger = logging.getLogger("slac_utils")

# commands
SIOCGIFHWADDR = 0x8927  # Get hardware address
SIOCGIFADDR = 0x8915  # get PA address

# From net/if_arp.h
ARPHDR_ETHER = 1
ARPHDR_METRICOM = 23
ARPHDR_PPP = 512
ARPHDR_LOOPBACK = 772
ARPHDR_TUN = 65534

NID_LENGTH = 7


def generate_nid(nmk: bytes):
    """
    It generates a NID key, based on the NMK, which is a random 16 bytes value.
    The way the NID is generated is by hashing recursively the NMK 5 times,
    reinitializing the sha256 buffer each time. It then collects the first
    7 bytes of the result and shifts 4 times the least significant byte.
    This algorithm was extracted from the Qualcomm implementation of it
    in https://github.com/qca/open-plc-utils/blob/master/key/HPAVKeyNID.c

    The procedure to get the NID described in the HPGP 1.1, section 4.4.3.1
    does not match the algorithm presented here and in fact makes no sense;
    During the testival in Stuttgart between 28-30 September 2021, tried to get
    someone to explain the logic described in HPGP but no one could.
    :param nmk: NMK [16 bytes] randomly generated
    :return: NID [7 bytes]
    """
    # For the sake of clarity, the attribute for the method is nmk, but over
    # the loop we want to keep updating the same variable and it becomes a
    # digest, so we set the initial digest variable value as equal to nmk
    digest = nmk
    for _ in range(5):
        _sha256 = sha256()
        _sha256.update(digest)
        digest = _sha256.digest()
    truncated_digest = digest[:NID_LENGTH]
    last_byte = NID_LENGTH - 1
    nid = truncated_digest[:last_byte] + (truncated_digest[last_byte] >> 4).to_bytes(
        1, "big"
    )
    return nid


def half_round(x):
    """
    Python round() function rounds the float number to the nearest even number,
    i.e:
    >> round(2.3)
       2
    >> round(2.5)
       2

    This happens because python 3.x in contrast to 2.x, uses Banker's rounding
    for the function (http://en.wikipedia.org/wiki/Banker's_rounding;
    https://wiki.c2.com/?BankersRounding)

    This function implements a half-way rounding, which is more commonly used,
    for positive and negative numbers. The copysign(0.5, x) provides the sign
    of the float number to be rounded and injects it in 0.5.

    """
    return int(x + copysign(0.5, x))


def time_now_ms():
    return round(time.time() * 1000)


def is_distro_linux() -> bool:
    """
    Checks if the Machine is a Linux one
    :return:
    """
    if platform.startswith("linux"):
        return True
    return False


def plain_str(x):
    """Convert basic byte objects to str"""
    if isinstance(x, bytes):
        return x.decode(errors="backslashreplace")
    return str(x)


def str2mac(s):
    """
    Returns the Mac Address in the format 00:00:00:00:00:00
    """
    if isinstance(s, str):
        return ("%02x:" * 6)[:-1] % tuple(map(ord, s))
    return ("%02x:" * 6)[:-1] % tuple(s)


def get_if_hwaddr(iff: str, to_mac_fmt=False):
    """
    Returns the MAC (hardware) address of an interface in readable format
    iff: interface name, e.g. en0 or enp0s3
    """

    addr_family, mac = get_if_raw_hwaddr_linux(iff)  # type: ignore # noqa: F405

    if addr_family in [ARPHDR_ETHER, ARPHDR_LOOPBACK]:
        if to_mac_fmt:
            return str2mac(mac)
        return mac
    raise Exception(
        f"Unsupported address family ({addr_family}) " "for interface [{iff}]"
    )


def get_if_raw_hwaddr_linux(iff: str, siocgifhwaddr: int = SIOCGIFHWADDR):
    """Get the raw MAC address of a local interface.
    This function uses SIOCGIFHWADDR (System I/O Command Get Interface HW Addr)
    calls, therefore only works
    on some distros.

    iff: str - Interface Designation (e.g. 'en0')
    siocgifhwaddr: int = SIOCGIFHWADDR - SIO command
    """

    sck = socket.socket()
    try:
        # input/output control
        raw_addr = ioctl(sck, siocgifhwaddr, struct.pack("16s16x", iff.encode("utf8")))
    except OSError as e:
        raise OSError(f"There is no such interface {iff}") from e
    finally:
        sck.close()
    # raw_addr is a 32 bytes buffer
    # 16 bytes are padding ones, thus they are ignored (16x)
    # we expect a short (h) - 2 bytes
    # a 6-bytes value that is converted as string (6s), like b'\x00\x01..\x06'
    # and 8 bytes are ignored as padding (8x)
    # so the value returned is a tuple with two items, like:
    # (1, b'\x02B\xac\x14\x00\x02')
    return struct.unpack("16xh6s8x", raw_addr)


async def cancel_task(task):
    """Cancel the task safely"""
    task.cancel()
    try:
        await task
    except asyncio.CancelledError:
        pass


def task_callback(task: asyncio.Task):
    """
    Callback for a task
    This is very useful when spawning tasks in the background with asyncio.create_task.
    As the task runs in the background any runtime exception is not logged, so with this
    callback, it is possible to keep logging the exceptions
    https://stackoverflow.com/questions/66293545/asyncio-re-raise-exception-from-a-task
    """
    try:
        task.result()
    except asyncio.CancelledError:
        pass  # Task cancellation should not be logged as an error.
    except Exception as e:
        logger.error(f"Exception raised by task: {task.get_name()}", e)


async def wait_till_finished(
    awaitables: List[Awaitable[Any]], finished_when=asyncio.FIRST_EXCEPTION
):
    # pylint: disable=broad-except
    """Run the tasks until one task is finished. The condition to finish
    depends on the argument 'finished_when', which directly translates
    to the asyncio.wait argument 'return_when' that can assume the following
    values: FIRST_COMPLETED, FIRST_EXCEPTION, ALL_COMPLETED
    (For more information regarding this, please check:
    https://docs.python.org/3/library/asyncio-task.html#waiting-primitives)

    All unfinished tasks will be cancelled.

    It can happen that multiple tasks finished at the same time.
    A MultiError is raised if at least one task finished with an exception.
    This exception wraps the exception of all tasks that finished with an
    exception.

    Return values of finished tasks are ignored. Use `asyncio.wait()` directly
    if you need access to the return values of tasks.

    If this function turns out to be useful it might be a good fit for
    `common/util` or `cc_utils`.

    """
    tasks = []

    # As of Python 3.8 `asyncio.wait()` should be called only with
    # `asyncio.Task`s.
    # See: https://docs.python.org/3/library/asyncio-task.html#asyncio-example-wait-coroutine  # noqa: E501
    for awaitable in awaitables:
        if not isinstance(awaitable, asyncio.Task):
            awaitable = asyncio.create_task(awaitable)
        tasks.append(awaitable)

    done, pending = await asyncio.wait(tasks, return_when=finished_when)

    for task in pending:
        await cancel_task(task)

    errors = []
    for task in done:
        try:
            task.result()
        except Exception as ex:
            logger.exception(ex)
            errors.append(ex)

    if len(errors) == 1:
        raise errors[0]

    if errors:
        raise MultiError(errors)


class MultiError(Exception):
    """Exception used to raise multiple exceptions.

    The attribute `errors` gives access to the wrapper errors.

        try:
            something()
        except MultiError as e:
            for error in e.errors:
                if isinstance(e, ZeroDivisionError):
                    ...
                elif isinstance(e, AttributeError):
                    ...

    """

    def __init__(self, errors: List[Exception]):
        Exception.__init__(self)
        self.errors = errors



from enum import Enum

# [V2G3-M06-05]- In case NO communication could be established with a
# 5 % control pilot duty cycle (matching process not started), if the EVSE
# wants to switch to a nominal duty cycle, then the change from 5 % to a
# nominal duty cycle shall be done with a specific sequence
# B2 or C2 (5 %) -> E or F -> B2 (nominal value) to allow backward
# compatibility. The minimum time at the control pilot state E or F is
# defined by T_step_EF.

# [V2G3-M06-06] In case a communication has already been established within
# 5 % control pilot duty cycle (“Matched state” reached or
# matching process ongoing), a change from 5 % to a nominal duty cycle shall be
# done with a X1 state in the middle (minimum time as defined in [IEC-3]
# Seq 9.2), to signal the EV that the control pilot duty cycle will change to a
# nominal duty cycle

# [V2G3 M06-07] - If an AC EVSE applies a 5 % control pilot duty cycle,
# and the EVSE receives NO SLAC request within TT_EVSE_SLAC_init, the EVSE
# shall go to state E or F for T_step_EF (min 4 s), shall go back to 5 % duty
# cycle, and  shall reset the TT_EVSE_SLAC_init timeout before being ready to
# answer a matching request again. This sequence shall be retried
# C_sequ_retry times (2). At the end, without any reaction, the EVSE shall go
# to state X1

# [V2G3-M06-08] After positive EIM, if no matching process is running, the EVSE
# shall signal control pilot state E/F for T_step_EF, then signal control pilot
# state X1/X2 (nominal).

# [V2G3 -M06-09] If a control pilot state E/F -> Bx, Cx, Dx transition is used
# for triggering retries or legacy issues, the state E/F shall be at least
# T_step_EF.

# [V2G2-852] - If no positive authorization information from an EIM has been
# received the EVSE shall apply PWM equal to 5 %.

# [V2G2-853] - If an EVSE receives positive authorization information from an
# EIM the EVSE shall apply a nominal duty cycle.

# [V2G2-931] - The EVSE shall signal a PWM of 5 % or nominal duty cycle after
# sending the message AuthorizationRes.


class Timers(float, Enum):
    """
    Timeouts defined by ISO15118-3 in table A.1
    All times are in seconds
    """

    # Time between the moment the EVSE detects state B and the reception of the
    # first SLAC Message, i.e. CM_SLAC_PARM.REQ.
    # This Timer is actually set in the environment.py, for debugging and
    # development reasons, allowing a easier setting of the time with the
    # docker-compose.dev.yml
    SLAC_INIT_TIMEOUT = 50.0  # [TT_EVSE_SLAC_init=20 s - 50 s]

    # Timeout for the reception of either CM_VALIDATE.REQ or CM_SLAC_MATCH.REQ
    # message, after reception of CM_ATTEN_CHAR.RSP
    SLAC_MATCH_TIMEOUT = 10.0  # [TT_EVSE_match_session=10 s]

    # Time the EV shall wait for CM_ATTEN_CHAR.IND after sending the first
    # CM_START_ATTEN_CHAR.IND
    SLAC_ATTEN_RESULTS_TIMEOUT = 1.2  # [TT_EV_atten_results = 1200 ms]

    # Timeout used for awaiting for a Request
    SLAC_REQ_TIMEOUT = 0.4  # [TT_match_sequence = 400 ms]

    # Timeout used for awaiting for a Response
    SLAC_RESP_TIMEOUT = 0.2  # [TT_match_response = 200 ms]

    # According to the standard:
    # [V2G3-A09-124] - In case the matching process is considered as FAILED,
    # wait for a time of TT_ matching_rate before restarting the process.

    # [V2G3-A09-125] - If the matching process fails for all retries started
    # within TT_matching_repetition, the matching process shall be stopped
    # in “Unmatched” state (see Figure 11).

    # The number maximum of retries is defined by C_conn_max_match = min 3
    # So, if within the TT_matching_repetition (10 s) time, the number of
    # retries expires, the matching process shall be stopped
    # in “Unmatched” state. (ISO Requirement couldnt be found for this,
    # but this is the logical steps to do)

    # Total time while the new SLAC repetitions can happen.
    # Once this timer is expired, the Matching process is considered FAILED
    SLAC_TOTAL_REPETITIONS_TIMEOUT = 10.0  # [TT_matching_repetition = 10 s]

    # Time to wait for the repetition of the matching process
    SLAC_REPETITION_TIMEOUT = 0.4  # [TT_matching_rate = 400 ms]

    # Time required to await while in state E or F (used in some use cases,
    # like the one defined by [V2G3 M06-07])
    SLAC_E_F_TIMEOUT = 4.0  # [T_step_EF = min 4 s]


# Timeout on the EVSE side that triggers the calculation of the average
# attenuation profile. Time is in multiples of 100 ms. In this case, we have
# 600 ms (6 * 100)

# Timers.SLAC_ATTEN_TIMEOUT is the only value in the Timers class that its use
# is supposed to be done as an integer (type int)
SLAC_ATTEN_TIMEOUT = 6  # [TT_EVSE_match_MNBC = 600 ms]
ATTEN_RESULTS_TIMEOUT = 1150  # in ms


class FramesSizes(int, Enum):
    """
    Frames Sizes in bytes of several messages
    (currently only EVSE side relevant ones)

    The size is calculated as the summation of the size of:
    * EthernetHeader = 14 bytes
    * HomePlugHeader = 5 bytes
    * MessageSize = X bytes

    For example, a complete CM_ATTEN_CHAR.RSP frame must have 70 Bytes:
    EthernetHeader = 14 bytes
    HomePlugHeader  = 5 bytes
    AttenCharRsp = 51 bytes

    CM_MNBC_SOUND.IND and CM_ATTEN_PROFILE.IND are not included
    """

    CM_SET_KEY_CNF = 60
    CM_SLAC_PARM_REQ = 60
    CM_START_ATTEN_CHAR_IND = 60
    CM_MNBC_SOUND_IND = 71
    CM_ATTEN_PROFILE_IND = 85
    CM_ATTEN_CHAR_RSP = 70
    CM_SLAC_MATCH_REQ = 85
    LINK_STATUS_CNF = 60


# Socket Receive buffer max frame is equal to the MAX ETH Frame Size
BUFF_MAX_SIZE = 1500
SLAC_RUNID_LEN = 8
# NumberOfSounds
SLAC_MSOUNDS = 10

# 15118-3 mentions the RESP_TYPE shall be 0x01 indicating other GP Stations
# However, qualcomm example uses 0x00 which indicates the recipient of the
# SOUND MPDUs shall communicate the signal attenuation characteristic profile
# data to the HLE instead of another GP STA.
SLAC_RESP_TYPE = 0x01
SLAC_APPLICATION_TYPE = 0x00
SLAC_SECURITY_TYPE = 0x00

# Pause used between sounds sent by the ev during cm_mnbc_sound routine
# By the spec can be a value between 20 and 50 ms (TP_EV_batch_msg_interval)
SLAC_PAUSE = 0.02
SLAC_GROUPS = 58
SLAC_LIMIT = 40
# Time to await after reception of a successful CM_SET_KEY.CNF
# This timer is used and defined in the Qualcomm example
SLAC_SETTLE_TIME = 10

ETHER_ADDR_LEN = 6
BROADCAST_ADDR = b"\xFF" * 6


CM_SET_KEY_TYPE = b"\x01"
# According to 15118-3 this value should be 0x00 and not 0xAA
CM_SET_KEY_MY_NONCE = b"\xaa\xaa\xaa\xaa"
# CM_SET_KEY_MY_NONCE = b'\x00' * 4
CM_SET_KEY_YOUR_NONCE = b"\x00\x00\x00\x00"
CM_SET_KEY_PID = b"\x04"
CM_SET_KEY_PRN = b"\x00\x00"
CM_SET_KEY_PMN = b"\x00"
CM_SET_KEY_NEW_EKS = b"\x01"
CM_SET_CCO_CAPAB = b"\x00"

# MMTypes Base Codes
CM_SET_KEY = 0x6008  # Equal to 24584 in decimal
CM_SLAC_PARM = 0x6064
CM_START_ATTEN_CHAR = 0x6068
CM_MNBC_SOUND = 0x6074
CM_ATTEN_PROFILE = 0x6084
CM_ATTEN_CHAR = 0x606C
CM_SLAC_MATCH = 0x607C

# MMType Kind
MMTYPE_REQ = 0x0000
MMTYPE_CNF = 0x0001
MMTYPE_IND = 0x0002
MMTYPE_RSP = 0x0003

# Ehternet Type HomePlug AV
ETH_TYPE_HPAV = 0x88E1

HOMEPLUG_MMV = b"\x01"
HOMEPLUG_FMSN = b"\x00"
HOMEPLUG_FMID = b"\x00"


# STATES
STATE_UNMATCHED = 0
STATE_MATCHING = 1
STATE_MATCHED = 2


# Station Identifier
EVSE_ID = "BBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBB"
# The dest MAC was defined in channel.c as follows in Qualcomm open-plc
EVSE_PLC_MAC = b"\x00\xb0\x52\x00\x00\x01"

# Qualcomm settings
# HomePlugAV0123 (defined in evse.c and also evse.ini of Qualcomm open-plc)
QUALCOMM_NID = b"\x02\x6b\xcb\xa5\x35\x4e\x08"
QUALCOMM_NMK = b"\xb5\x93\x19\xd7\xe8\x15\x7b\xa0\x01\xb0\x18\x66\x9c\xce\xe3\x0d"



from pyslac.utils import is_distro_linux

if not is_distro_linux():
    raise EnvironmentError("Non-Linux systems are not supported")

import asyncio
import json
import logging
import os
from typing import List, Optional

from pyslac.environment import Config
from pyslac.session import SlacEvseSession, SlacSessionController
from pyslac.utils import wait_till_finished

logging.basicConfig(level=logging.DEBUG)
logger = logging.getLogger(__file__)


class SlacHandler(SlacSessionController):
    def __init__(self, slac_config: Config):
        SlacSessionController.__init__(self)
        self.slac_config = slac_config
        self.running_sessions: List["SlacEvseSession"] = []

    async def notify_matching_ongoing(self, evse_id: str):
        """overrides the notify matching ongoing method defined in
        SlacSessionController"""
        logger.info(f"Matching is ongoing for {evse_id}")

    async def enable_hlc_charging(self, evse_id: str):
        """
        overrides the enable_hlc_charging method defined in SlacSessionController
        """
        logger.info(f"Enable PWM and set 5% duty cycle for evse {evse_id}")

    async def start(self, cs_config: dict):
        if cs_config["number_of_evses"] < 1 or (
            len(cs_config["parameters"]) != cs_config["number_of_evses"]
        ):
            raise AttributeError("Number of evses provided is invalid.")

        evse_params: dict = cs_config["parameters"][0]
        evse_id: str = evse_params["evse_id"]
        network_interface: str = evse_params["network_interface"]
        try:
            slac_session = SlacEvseSession(evse_id, network_interface, self.slac_config)
            await slac_session.evse_set_key()
            self.running_sessions.append(slac_session)
        except (OSError, TimeoutError, ValueError) as e:
            logger.error(
                f"PLC chip initialization failed for "
                f"EVSE {evse_id}, interface "
                f"{network_interface}: {e}. \n"
                f"Please check your settings."
            )
            return

        await self.enable_hlc_and_trigger_slac(self.running_sessions[0])

    async def enable_hlc_and_trigger_slac(self, session):
        """
        Dummy method to fake the enabling of the HLC by setting PWM to 5%
        and triggers the Matching by handling a CP state change to "B"
        """
        await self.enable_hlc_charging(session.evse_id)
        await self.process_cp_state(session, "B")
        await asyncio.sleep(2)
        await self.process_cp_state(session, "C")
        await asyncio.sleep(20)
        await self.process_cp_state(session, "A")


async def main(env_path: Optional[str] = None):
    # get configuration
    slac_config = Config()
    slac_config.load_envs(env_path)
    root_dir = os.path.dirname(os.path.abspath(__file__))
    json_file = open(os.path.join(root_dir, "cs_configuration.json"))
    cs_config = json.load(json_file)
    json_file.close()
    slac_handler = SlacHandler(slac_config)
    tasks = [slac_handler.start(cs_config)]
    await wait_till_finished(tasks)


def run():
    asyncio.run(main())


if __name__ == "__main__":
    run()



import logging
import os
from dataclasses import dataclass
from typing import Optional

import environs

from pyslac.enums import Timers

logger = logging.getLogger(__name__)


@dataclass
class Config:
    slac_init_timeout: Optional[int] = None
    log_level: Optional[int] = None

    def load_envs(self, env_path: Optional[str] = None) -> None:
        """
        Tries to load the .env file containing all the project settings.
        If `env_path` is not specified, it will get the .env on the current
        working directory of the project
        Args:
            env_path (str): Absolute path to the location of the .env file
        """
        env = environs.Env(eager=False)
        if not env_path:
            env_path = os.getcwd() + "/.env"
        env.read_env(path=env_path)  # read .env file, if it exists

        # This timer is set in docker-compose.dev.yml, for merely debugging and dev
        # reasons
        self.slac_init_timeout = env.float(
            "SLAC_INIT_TIMEOUT", default=Timers.SLAC_INIT_TIMEOUT
        )

        self.log_level = env.str("LOG_LEVEL", default="INFO")

        env.seal()  # raise all errors at once, if any



"""
from https://github.com/spotify/linux/blob/master/include/linux/if_ether.h
"""
import ctypes

#  BPF FILTERS ENUMS
# Instruction classes
BPF_LD = 0x00
BPF_JMP = 0x05
BPF_RET = 0x06

# ld/ldx fields
BPF_H = 0x08
BPF_B = 0x10
BPF_ABS = 0x20

# alu/jmp fields
BPF_JEQ = 0x10
BPF_K = 0x00

# As defined in asm/socket.h
SO_ATTACH_FILTER = 26


# ETH TYPE ENUMS
# Dummy type for 802.3 frames
ETH_P_802_3 = 0x0001
# Dummy protocol id for AX.25
ETH_P_AX25 = 0x0002
# Every packet (be careful!!!)
ETH_P_ALL = 0x0003
# Internet Protocol Packet
ETH_P_IP = 0x0800
# HomePlug AV Protocol Packet
ETH_P_HPAV = 0x88E1


# Used to set Promiscuous mode in Linux
# https://stackoverflow.com/questions/6067405/python-sockets-enabling-promiscuous-mode-in-linux/6072625
IFF_PROMISC = 0x100
SIOCGIFFLAGS = 0x8913
SIOCSIFFLAGS = 0x8914


class ifreq(ctypes.Structure):
    _fields_ = [("ifr_ifrn", ctypes.c_char * 16), ("ifr_flags", ctypes.c_short)]



import ctypes
from dataclasses import dataclass
from typing import List

from pyslac.enums import (
    BROADCAST_ADDR,
    CM_SET_CCO_CAPAB,
    CM_SET_KEY_MY_NONCE,
    CM_SET_KEY_NEW_EKS,
    CM_SET_KEY_PID,
    CM_SET_KEY_PMN,
    CM_SET_KEY_PRN,
    CM_SET_KEY_TYPE,
    CM_SET_KEY_YOUR_NONCE,
    SLAC_APPLICATION_TYPE,
    SLAC_ATTEN_TIMEOUT,
    SLAC_MSOUNDS,
    SLAC_RESP_TYPE,
    SLAC_SECURITY_TYPE,
)


@dataclass
class SetKeyReq:
    """
    Associated with CM_SET_KEY.REQ, defined in chapter 11.5.4 of the HPGP
    standard. Check also table page 586, table 11-87

    Also table A.8 from ISO15118-3

    EVSE/PEV -> HPGP Node

    This payload is defined as follows:
    |KeyType|MyNonce|YourNonce|PID|PRN|PMN|CCoCap|NID|NewEKS|NewKey|

    KeyType [1 byte] = 0x01: Fixed value to indicate NMK
    MyNonce [4 bytes] = 0x00000000: Fixed value, used by the emitter of the
                                    message and fixed over one session.
                                    If in another message, the receiver receives
                                    a different value, then it may consider
                                    that the communication was compromised
    YourNonce [4 bytes] = 0x00000000: Fixed value, encrypted payload not used.
                                      This field has the same rationale as
                                      MyNonce but in the opposite direction
                                      of the communication
    PID [1 byte] = 0x04: Fixed value to indicate "HLE protocol"
    PRN [2 bytes] = 0x0000: Fixed value, encrypted payload not used
    PMN [1 byte] = 0x00: Fixed value, encrypted payload not used
    CCo Capability [1 byte] = 0x00 : CCo Capability according to station role;
                                     Ususally the value is variable, but is used
    NID [7 bytes]: 54 LSBs contain the Network identifier and the rest is 0b00
                   Network ID derived from the NMK by the EVSE according to
                   [HPGP], 4.4.3.1
    NewEKS [1 byte] = 0x01: Fixed value to indicate NMK

    NewKey [16 bytes]: NMK (Network Mask, random value per session)

    Message size is = 44 bytes
    """

    # 7bytes
    nid: bytes
    # 16 bytes
    new_key: bytes

    def __bytes__(self, endianess: str = "big"):
        if endianess == "big":
            return (
                CM_SET_KEY_TYPE
                + CM_SET_KEY_MY_NONCE
                + CM_SET_KEY_YOUR_NONCE
                + CM_SET_KEY_PID
                + CM_SET_KEY_PRN
                + CM_SET_KEY_PMN
                + CM_SET_CCO_CAPAB
                + self.nid
                + CM_SET_KEY_NEW_EKS
                + self.new_key
            )
        return (
            self.new_key
            + CM_SET_KEY_NEW_EKS
            + self.nid
            + CM_SET_CCO_CAPAB
            + CM_SET_KEY_PMN
            + CM_SET_KEY_PRN
            + CM_SET_KEY_PID
            + CM_SET_KEY_YOUR_NONCE
            + CM_SET_KEY_MY_NONCE
            + CM_SET_KEY_TYPE
        )

    def pack_big(self):
        return self.__bytes__()

    def pack_little(self):
        return self.__bytes__("little")


@dataclass
class SetKeyCnf:
    """
    Associated with CM_SET_KEY.CNF, defined in chapter 11.5.5 of the HPGP
    standard. Check also table page 586, table 11-87

    Also table A.8 from ISO15118-3

    HPGP Node -> EVSE/PEV

    This payload is defined as follows:
    |Result|MyNonce|YourNonce|PID|PRN|PMN|CCoCap|

    Result [1 byte]: 0x00 - Success, 0x01 - Failure
    MyNonce [4 bytes]: Random number that will be used to verify next message
                       from other end; in encrypted portion of payload.
    YourNonce [4 bytes]: Last nonce received from recipient; it will be used
                         by recipient to verify this message;
                         in encrypted portion of payload.
    PID [1 byte]: Protocol for which Set Key is confirmed
    PRN [2 bytes]: Protocol Run Number (refer to Section 11.5.2.4)
    PMN [1 byte]: Protocol Message Number (refer to Section 11.5.2.5
    CCo Capability [1 byte]: The two LSBs of this field contain the STA’s
                             CCo capability. The interpretation of these bits
                             is the same as in Section 4.4.3.13.4.6.2.
                             The six MSBs of this field are set to 0b000000

    Message size is = 14 bytes
    """

    result: int
    my_nonce: bytes
    your_nonce: bytes
    pid: bytes
    prn: bytes
    pmn: bytes
    cco_capab: bytes

    def __bytes__(self, endianess: str = "big"):
        frame = bytearray(
            self.result.to_bytes(1, "big")
            + self.my_nonce
            + self.your_nonce
            + self.pid
            + self.prn
            + self.pmn
            + self.cco_capab
        )
        if endianess == "big":
            return frame
        return frame.reverse()

    def pack_big(self):
        return self.__bytes__()

    def pack_little(self):
        return self.__bytes__("little")

    @classmethod
    def from_bytes(cls, payload: ctypes) -> "SetKeyCnf":
        # payload: Union[int, bytes]
        # self.result = ctypes.c_uint8(payload[19])
        # The Ethernet and HP header comprise 19 bytes in total
        # result = payload[19]
        #  TODO: Think about pair this with the Homeplug Greenphy Header and
        # check the MMV == HOMEPLUG_MMV and
        # MMType == (CM_SET_KEY | MMTYPE_CNF) fields
        # The result is ignored, because some Qualcomm chips firmware have
        # implemented the wrong logic for it and a result of 0x00 means a
        # failure. In the future we can check if the SET_KEY_REQ was successful
        # by sending a NW_INFO.REQ to check if the NID was really set
        # if result != 0x00:
        #     TODO: Raise SLAC Exception
        #     raise ValueError("Device refused SET_KEY_REQ ")
        return cls(
            result=payload[19],
            my_nonce=payload[20:24],
            your_nonce=payload[24:28],
            pid=payload[28],
            prn=payload[29:31],
            pmn=payload[31],
            cco_capab=payload[32],
        )


@dataclass
class SlacParmReq:
    """
    Broadcast Message
    PEV -> EVSE

    Associated with CM_SLAC_PARM.REQ, defined in chapter 11.5.45 of the HPGP
    standard. Check also table page 586, table 11-87

    Also table A.2 from ISO15118-3

    This payload is defined as follows originally :
    |Application Type|Security Type|Run ID|CipherSuiteSetSize| CipherSuite..

    Application Type [1 byte]: 0x00 Fixed value indicating 'PEV- EVSE matching'
    Security Type [1 bytes]: 0x00 Fixed value indicating 'No Security'
    Run ID [8 bytes]: Identifier for a matching run, randomly chosen by
                      the EV for each CM_SLAC_PARM.REQ message and constant
                      for all following messages of the same run
    CipherSuiteSetSize [1 byte]: Number of supported cipher suites N.
    CipherSuite [1] [2 bytes]: First supported cipher suite.
    CipherSuite [N] [2 bytes]: Nth supported cipher suite.

    However, since Security Type is set as 0x00, Cipher suit is not used, thus
    the payload resumes to:
    |Application Type|Security Type|Run ID|

    Message size is = 10 bytes
    """

    # 8 bytes
    run_id: bytes
    application_type: int = SLAC_APPLICATION_TYPE
    security_type: int = SLAC_SECURITY_TYPE

    def __bytes__(self, endianess: str = "big"):
        frame = bytearray(
            self.application_type.to_bytes(1, "big")
            + self.security_type.to_bytes(1, "big")
            + self.run_id
        )
        if endianess == "big":
            return frame
        return frame.reverse()

    def pack_big(self):
        return self.__bytes__()

    def pack_little(self):
        return self.__bytes__("little")

    @classmethod
    def from_bytes(cls, payload: ctypes) -> "SlacParmReq":
        return cls(
            application_type=payload[19],
            security_type=payload[20],
            run_id=payload[21:29],
        )


@dataclass
class SlacParmCnf:
    # pylint: disable=too-many-instance-attributes
    """
    Unicast Message
    EVSE -> PEV

    PEV-HLE broadcasts CM_SLAC_PARM requests until at least one
    matching CM_SLAC_PARM confirm is received; matching confirm
    has the same run identifier; EVSE-HLE returns information to
    the PEV-HLE;

    Associated with CM_SLAC_PARM.CNF, defined in chapter 11.5.46 of the HPGP
    standard. Check also table page 586, table 11-87

    Also table A.2 from ISO15118-3

    This payload is defined as follows, originally :
    | M-SOUND_TARGET | NUM_SOUNDS| Time_Out| RESP_TYPE |
    |FORWARDING_STA | APPLICATION_TYPE| SECURITY_TYPE| RunID| *CipherSuite*

    M-SOUND_TARGET [6 byte] - 0xFFFFFFFFFFFF: Indicates the MAC address of the
                                              GP STA with which the STA shall
                                              initiate the Signal Level
                                              Attenuation Characterization
                                              Process.Fixed value indicating
                                              that M-Sounds to be sent as
                                              Ethernet broadcast

    NUM_SOUNDS [1 byte] - SLAC_MSOUNDS: Number of expected M-Sounds
                                        transmitted by the GP station
                                        during the SLAC process
    Time_Out [1 byte] - SLAC_ATTEN_TIMEOUT: Duration TT_EVSE_match_MNBC while
                                          the EVSE receives incoming M-SOUNDS
                                          after a CM_START_ATTEN_CHAR.IND. On
                                          other words, indicates the amount of
                                          time within which the GP STA will
                                          complete the transmission of SOUND
                                          MPDUs during the Signal Level
                                          Attenuation Characterization Process.
                                          The time is in multiples of 100 msec.
                                          E.g, Time_Out = 6 corresponds to 600ms

    RESP_TYPE [1 byte] - SLAC_RESP_TYPE: Fixed value indicating 'Other GP station'
                                         Indicates whether the recipient of the
                                         SOUND MPDUs shall communicate the signal
                                         attenuation characteristic profile data to
                                         the HLE or another GP STA.
    FORWARDING_STA [6 bytes]: EV Host MAC; The destination of SLAC results is
                              always the EV Host. Only valid if RESP_TYPE = 0x01
    APPLICATION_TYPE [1 byte] - 0x00: Fixed value indicating 'PEV-EVSE Matching'
    SECURITY_TYPE [1 byte] - 0x00: Fixed value indicating “No Security”
    RunID [8 bytes]: This value shall be the same as the one sent in the
                     CM_SLAC_PARM.REQ message by the EV
    * CipherSuite [2 bytes] *: Selected Cipher Suite

    *Since Security Type is 0x00, CipherSuite wont be present in the payload

    Message size is = 25 bytes
    """
    # 6 bytes
    forwarding_sta: bytes
    # 8 bytes
    run_id: bytes
    msound_target: bytes = BROADCAST_ADDR
    num_sounds: int = SLAC_MSOUNDS
    time_out: int = SLAC_ATTEN_TIMEOUT
    resp_type: int = SLAC_RESP_TYPE
    application_type: int = SLAC_APPLICATION_TYPE
    security_type: int = SLAC_SECURITY_TYPE

    def __bytes__(self, endianess: str = "big"):
        frame = bytearray(
            self.msound_target
            + self.num_sounds.to_bytes(1, "big")
            + self.time_out.to_bytes(1, "big")
            + self.resp_type.to_bytes(1, "big")
            + self.forwarding_sta
            + self.application_type.to_bytes(1, "big")
            + self.security_type.to_bytes(1, "big")
            + self.run_id
        )
        if endianess == "big":
            return frame
        return frame.reverse()

    def pack_big(self):
        return self.__bytes__()

    def pack_little(self):
        return self.__bytes__("little")

    @classmethod
    def from_bytes(cls, payload: ctypes) -> "SlacParmCnf":

        return cls(
            msound_target=payload[19:25],
            num_sounds=payload[25],
            time_out=payload[26],
            resp_type=payload[27],
            forwarding_sta=payload[28:34],
            application_type=payload[34],
            security_type=payload[35],
            run_id=payload[36:44],
        )


@dataclass
class StartAtennChar:
    """
    Broadcast Message

    PEV -> EVSE

    Associated with CM_START_ATTEN_CHAR.IND in chapter 11.5.47 of the HPGP
    standard. Check also table page 586, table 11-87

    Also table A.4 from ISO15118-3

    This payload is defined as follows originally :

    |Application Type|Security Type| NUM_SOUNDS| Time_Out| RESP_TYPE |
    |FORWARDING_STA |RunID|


    Application Type [1 byte]: 0x00 Fixed value indicating 'PEV- EVSE matching'
    Security Type [1 bytes]: 0x00 Fixed value indicating 'No Security'

    The following parameters are under a nested field called in the HPGP
    standard as ACVarField (Attenuation Characterization Variable Field)

    NUM_SOUNDS [1 byte]: Number of M-Sounds transmitted by the GP station
                         during the SLAC process
    Time_Out [1 byte]: Max time window on which the M-Sounds are sent,
                       associated with TT_EVSE_match_MNBC of the spec and
                       SLAC_ATTEN_TIMEOUT of enums.py. Multiple of 100 ms, i.e.,
                       if Time_Out = 6, it means the timeout is equal to 600ms
    RESP_TYPE [1 byte] - SLAC_RESP_TYPE: Fixed value indicating
                                         'Other Green PHY station'
    FORWARDING_STA [6 bytes]: EV Host MAC; The destination of SLAC results is
                              always the EV Host
    RunID [8 bytes]: This value shall be the same as the one in
                     CM_SLAC_PARM.REQ message sent by the EV

    Message size is = 19 bytes
    """

    num_sounds: int
    time_out: int
    # 6 bytes
    forwarding_sta: bytes
    # 8 bytes
    run_id: bytes
    application_type: int = SLAC_APPLICATION_TYPE
    security_type: int = SLAC_SECURITY_TYPE
    resp_type: int = SLAC_RESP_TYPE

    def __bytes__(self, endianess: str = "big"):
        frame = bytearray(
            self.application_type.to_bytes(1, "big")
            + self.security_type.to_bytes(1, "big")
            + self.num_sounds.to_bytes(1, "big")
            + self.time_out.to_bytes(1, "big")
            + self.resp_type.to_bytes(1, "big")
            + self.forwarding_sta
            + self.run_id
        )
        if endianess == "big":
            return frame
        return frame.reverse()

    def pack_big(self):
        return self.__bytes__()

    def pack_little(self):
        return self.__bytes__("little")

    @classmethod
    def from_bytes(cls, payload: ctypes) -> "StartAtennChar":
        return cls(
            application_type=payload[19],
            security_type=payload[20],
            num_sounds=payload[21],
            time_out=payload[22],
            resp_type=payload[23],
            forwarding_sta=payload[24:30],
            run_id=payload[30:38],
        )


@dataclass
class MnbcSound:
    """
    Broadcast Message

    PEV -> EVSE (HPGP Node)

    Associated with CM_MNBC_SOUND.IND in chapter 11.5.54 of the HPGP
    standard. Check also table page 586, table 11-87

    Also table A.4 from ISO15118-3

    This payload is defined as follows, originally :

    |Application Type|Security Type|SenderID|Cnt|RunID|RSVD|Rnd|


    Application Type [1 byte]: 0x00 Fixed value indicating 'PEV- EVSE matching'
    Security Type [1 bytes]: 0x00 Fixed value indicating 'No Security'

    The following parameters are under a nested field called in the HPGP
    standard as MSVarField (MNBC Sound Variable Field)

    SenderID [17 bytes] - 0x00: Sender’s Identification. According to HPGP:
                                If APPLICATION_TYPE=0x00 then Sender ID is
                                PEV’s VIN code.
                                But 15118-3 defines it as 0x00
    Cnt [1 byte]: Countdown counter for number of Sounds remaining

    According to HPGP spec, RunID has 16 Bytes, but 15118-3 just uses 8 bytes,
    so the other 8 are set to 0x00 and are reserved

    RunID [8 bytes]: This value shall be the same as the one in
                     CM_SLAC_PARM.REQ message sent by the EV
    RSVD [8 bytes] - 0x00: Reserved
    Rnd [16 bytes]: Random Value

    Message size is = 52 bytes
    """

    cnt: int
    # 8 bytes
    run_id: bytes
    application_type: int = SLAC_APPLICATION_TYPE
    security_type: int = SLAC_SECURITY_TYPE
    # 17 bytes
    sender_id: int = 0x00
    # 8 bytes
    rsvd: int = 0x00
    # 16 bytes
    rnd: int = 0xFF01

    def __bytes__(self, endianess: str = "big"):
        frame = bytearray(
            self.application_type.to_bytes(1, "big")
            + self.security_type.to_bytes(1, "big")
            + self.sender_id.to_bytes(17, "big")
            + self.cnt.to_bytes(1, "big")
            + self.run_id
            + self.rsvd.to_bytes(8, "big")
            + self.rnd.to_bytes(16, "big")
        )
        if endianess == "big":
            return frame
        return frame.reverse()

    def pack_big(self):
        return self.__bytes__()

    def pack_little(self):
        return self.__bytes__("little")

    @classmethod
    def from_bytes(cls, payload: ctypes) -> "MnbcSound":
        return cls(
            application_type=payload[19],
            security_type=payload[20],
            sender_id=int.from_bytes(payload[21:38], "big"),
            cnt=payload[38],
            run_id=payload[39:47],
            rsvd=int.from_bytes(payload[47:55], "big"),
            rnd=int.from_bytes(payload[55:71], "big"),
        )


@dataclass
class AttenProfile:
    """
    Sent by the HLE (HighLevel Entity/PLC chip) to the EVSE host application

    Associated with CM_ATTEN_PROFILE.IND

    Check table A.4 from ISO15118-3

    This payload is defined as follows originally :

    |PEV MAC|NumGroups|RSVD|AAG 1| AAG 2| AAG 3...|


    PEV MAC [6 byte]: MAC address of EV Host
    NumGroups [1 bytes]: 0x3A Number of OFDM carrier groups used for the SLAC
                              signal characterization.
    RSVD [1 bytes] - 0x00: Reserved
    AAG 1 [1 byte]: Average Attenuation of Group 1
    AAG Nth [1 byte]: Average Attenuation of Group Nth

    Message size is = 66 bytes
    """

    pev_mac: bytes
    # it has the length of num_groups bytes
    aag: List[int]
    # 0x3A = 58 Groups
    num_groups: int = 0x3A
    rsvd: int = 0x00

    def __bytes__(self, endianess: str = "big"):
        # I probably could use here
        # ``` python
        # bytearray(self.aag)
        # ```
        # but I wouldnt have
        # control over the number of groups used, so I decided here to go safe
        aag_bytes = b""
        for group in range(self.num_groups):
            aag_bytes += self.aag[group].to_bytes(1, "big")
        frame = bytearray(
            self.pev_mac
            + self.num_groups.to_bytes(1, "big")
            + self.rsvd.to_bytes(1, "big")
            + aag_bytes
        )
        if endianess == "big":
            return frame
        return frame.reverse()

    def pack_big(self):
        return self.__bytes__()

    def pack_little(self):
        return self.__bytes__("little")

    @classmethod
    def from_bytes(cls, payload: ctypes) -> "AttenProfile":
        num_groups = payload[25]
        return cls(
            pev_mac=payload[19:25],
            num_groups=num_groups,
            rsvd=payload[26],
            aag=list(payload[27 : 27 + num_groups]),
        )


@dataclass
class AtennChar:
    # pylint: disable=too-many-instance-attributes
    """
    Unicast Message

    Associated with CM_ATTEN_CHAR.IND in chapter 11.5.48 of the HPGP
    standard. Check also table page 586, table 11-87

    With this message, the EVSE shares with the PEV, the results of
    the sounds received per group.

    Also table A.4 from ISO15118-3

    EVSE -> PEV

    This payload is defined as follows originally :

    |Application Type|Security Type| SOURCE_ADDRESS| RunID| SOURCE_ID| RESP_ID|
    |NumSounds| ATTEN_PROFILE|

    ATTEN_PROFILE = |NumGroups|AAG 1| AAG 2| AAG 3...|

    Application Type [1 byte]: 0x00 Fixed value indicating 'PEV- EVSE matching'
    Security Type [1 bytes]: 0x00 Fixed value indicating 'No Security'

    The following parameters are under a nested field called in the HPGP
    standard as ACVarField (Attenuation Characterization Variable Field)

    SOURCE_ADDRESS [6 bytes]: MAC Address of EV Host which initiated the SLAC
    RunID [8 bytes]: This value shall be the same as the one in
                     CM_SLAC_PARM.REQ message sent by the EV
    SOURCE_ID [17 bytes] - 0x00...00: - The unique identifier of the station
                                        that sent the M-Sounds (not used by ISO)
    RESP_ID [17 bytes] - 0x00...00: - The unique identifier of the station that
                                      is sending this message (not used by ISO)
    NUM_SOUNDS [1 byte]: Number of M-Sounds used for generation of the
                         ATTEN_PROFILE
    ATTEN_PROFILE [59 bytes]: Signal Level Attenuation (Field format in table
                              'ATTEN_PROFILE' of [HPGP])
                              ATTEN_PROFILE = |NumGroups|AAG 1| AAG 2| AAG 3...|

    Message size is = 110 bytes
    """
    # 6 bytes
    source_address: bytes
    # 8 bytes
    run_id: bytes
    num_sounds: int
    num_groups: int
    #  255 bytes
    aag: List[int]

    application_type: int = SLAC_APPLICATION_TYPE
    security_type: int = SLAC_SECURITY_TYPE
    #  17 Bytes
    source_id: int = 0x00
    # 17 bytes
    resp_id: int = 0x00

    def __bytes__(self, endianess: str = "big"):
        frame = bytearray(
            self.application_type.to_bytes(1, "big")
            + self.security_type.to_bytes(1, "big")
            + self.source_address
            + self.run_id
            + self.source_id.to_bytes(17, "big")
            + self.resp_id.to_bytes(17, "big")
            + self.num_sounds.to_bytes(1, "big")
            + self.num_groups.to_bytes(1, "big")
            + bytearray(self.aag)
        )
        if endianess == "big":
            return frame
        return frame.reverse()

    def pack_big(self):
        return self.__bytes__()

    def pack_little(self):
        return self.__bytes__("little")

    @classmethod
    def from_bytes(cls, payload: ctypes) -> "AtennChar":
        num_groups = payload[70]
        return cls(
            application_type=payload[19],
            security_type=payload[20],
            source_address=payload[21:27],
            run_id=payload[27:35],
            source_id=int.from_bytes(payload[35:52], "big"),
            resp_id=int.from_bytes(payload[52:69], "big"),
            num_sounds=payload[69],
            num_groups=num_groups,
            aag=list(payload[71 : 71 + num_groups]),
        )


@dataclass
class AtennCharRsp:
    """
    Unicast Message

    Associated with CM_ATTEN_CHAR.RSP in chapter 11.5.49 of the HPGP
    standard. Check also table page 586, table 11-87

    Also table A.4 from ISO15118-3

    PEV -> EVSE

    This payload is defined as follows originally :

    |Application Type|Security Type| SOURCE_ADDRESS| RunID| SOURCE_ID| RESP_ID|
    |Result|


    Application Type [1 byte]: 0x00 Fixed value indicating 'PEV- EVSE matching'
    Security Type [1 bytes]: 0x00 Fixed value indicating 'No Security'

    The following parameters are under a nested field called in the HPGP
    standard as ACVarField (Attenuation Characterization Variable Field)

    SOURCE_ADDRESS [6 bytes]: MAC Address of EV Host which initiated the SLAC
    RunID [8 bytes]: This value shall be the same as the one in
                     CM_SLAC_PARM.REQ message sent by the EV
    SOURCE_ID [17 bytes] - 0x00: - HPGP defines it as the unique identifier
                                 from the station that sent the M-sounds.
                                 ISO15118-3 defines it as 0x00
    RESP_ID [17 byte] - 0x00: - HPGP defines it as the unique identifier
                               of the station that is sending this message.
                               ISO15118-3 defines it as 0x00
    Result [1 byte] - 0x00: Fixed value of 0x00 indicates a successful SLAC
                            process

    Message size is = 43 bytes
    """

    # 6 bytes
    source_address: bytes
    # 8 bytes
    run_id: bytes
    #  17 Bytes
    source_id: int
    # 17 bytes
    resp_id: int
    #  1 byte
    result: int

    application_type: int = SLAC_APPLICATION_TYPE
    security_type: int = SLAC_SECURITY_TYPE

    def __bytes__(self, endianess: str = "big"):
        frame = bytearray(
            self.application_type.to_bytes(1, "big")
            + self.security_type.to_bytes(1, "big")
            + self.source_address
            + self.run_id
            + self.source_id.to_bytes(17, "big")
            + self.resp_id.to_bytes(17, "big")
            + self.result.to_bytes(1, "big")
        )
        if endianess == "big":
            return frame
        return frame.reverse()

    def pack_big(self):
        return self.__bytes__()

    def pack_little(self):
        return self.__bytes__("little")

    @classmethod
    def from_bytes(cls, payload: ctypes) -> "AtennCharRsp":
        return cls(
            application_type=payload[19],
            security_type=payload[20],
            source_address=payload[21:27],
            run_id=payload[27:35],
            source_id=int.from_bytes(payload[35:52], "big"),
            resp_id=int.from_bytes(payload[52:69], "big"),
            result=payload[69],
        )


@dataclass
class MatchReq:
    # pylint: disable=too-many-instance-attributes
    """
    Unicast Message

    Associated with CM_SLAC_MATCH.REQ in chapter 11.5.57 of the HPGP
    standard. Check also table A.7 from ISO15118-3

    PEV -> EVSE

    This payload is defined as follows originally :

    |Application Type|Security Type| MVFLength| PEV_ID| PEV_MAC| EVSE_ID|
    EVSE MAC|RunID|RSVD|


    Application Type [1 byte] - 0x00: Fixed value indicating 'PEV- EVSE matching'
    Security Type [1 bytes] - 0x00: Fixed value indicating 'No Security'
    MVFLength [2 bytes] - 0x3e: (Fixed value) Match Variable Field Length

    The following parameters are under a nested field called in the HPGP
    standard as MatchVarField (Match Variable Field)

    PEV ID [17 bytes] - 0x00:
    PEV MAC [6 bytes]: MAC Address of the EV Host
    EVSE ID [17 bytes] - 0x00:
    EVSE MAC [6 bytes]: MAC Address of the EVSE Host
    RunID [8 bytes]: This value shall be the same as the one in
                     CM_SLAC_PARM.REQ message sent by the EV
    RSVD [8 bytes] - 0x00: Reserved

    Message size is = 62 bytes
    """
    # 6 bytes
    pev_mac: bytes
    # 6 bytes
    evse_mac: bytes
    # 8 bytes
    run_id: bytes
    application_type: int = SLAC_APPLICATION_TYPE
    security_type: int = SLAC_SECURITY_TYPE
    # 2 bytes
    mvf_length: int = 0x003E
    # 17 bytes
    pev_id: int = 0x00
    # 17 bytes
    evse_id: int = 0x00
    # 8 bytes
    rsvd: int = 0x00

    def __bytes__(self, endianess: str = "big"):
        frame = bytearray(
            self.application_type.to_bytes(1, "big")
            + self.security_type.to_bytes(1, "big")
            + self.mvf_length.to_bytes(2, "big")
            + self.pev_id.to_bytes(17, "big")
            + self.pev_mac
            + self.evse_id.to_bytes(17, "big")
            + self.evse_mac
            + self.run_id
            + self.rsvd.to_bytes(8, "big")
        )
        if endianess == "big":
            return frame
        return frame.reverse()

    def pack_big(self):
        return self.__bytes__()

    def pack_little(self):
        return self.__bytes__("little")

    @classmethod
    def from_bytes(cls, payload: ctypes) -> "MatchReq":
        return cls(
            application_type=payload[19],
            security_type=payload[20],
            mvf_length=int.from_bytes(payload[21:23], "big"),
            pev_id=int.from_bytes(payload[23:40], "big"),
            pev_mac=payload[40:46],
            evse_id=int.from_bytes(payload[46:63], "big"),
            evse_mac=payload[63:69],
            run_id=payload[69:77],
            rsvd=int.from_bytes(payload[77:85], "big"),
        )


@dataclass
class MatchCnf:
    # pylint: disable=too-many-instance-attributes
    """
    Unicast Message

    Associated with CM_SLAC_MATCH.CNF in chapter 11.5.58 of the HPGP
    standard. Check also table A.7 from ISO15118-3

    The selected EVSE responses to the EV request with a CM_SLAC_MATCH.CNF,
    which contains all parameters to be set to join the logical network of
    the EVSE.

    EVSE -> PEV

    This payload is defined as follows originally:

    |Application Type|Security Type| MVFLength| PEV_ID| PEV_MAC| EVSE_ID|
    EVSE MAC|RunID|RSVD1|NID|RSVD2|NMK


    Application Type [1 byte] - 0x00: Fixed value indicating 'PEV- EVSE matching'
    Security Type [1 bytes] - 0x00: Fixed value indicating 'No Security'
    MVFLength [2 bytes] - 0x56: (Fixed value) Match Variable Field Length

    The following parameters are under a nested field called in the HPGP
    standard as MatchVarField (Match Variable Field)

    PEV ID [17 bytes] - 0x00:
    PEV MAC [6 bytes]: MAC Address of the EV Host
    EVSE ID [17 bytes] - 0x00:
    EVSE MAC [6 bytes]: MAC Address of the EVSE Host
    RunID [8 bytes]: This value shall be the same as the one in
                     CM_SLAC_PARM.REQ message sent by the EV
    RSVD1 [8 bytes] - 0x00: Reserved
    NID [7 bytes]: Network ID derived from the NMK by the EVSE
                   according to [HPGP], 4.4.3.1
    RSVD2 [8 bytes] - 0x00: Reserved
    NMK [16 bytes]: Private Network Membership Key of the EVSE (random value)

    Message size is = 97 bytes
    """
    # 6 bytes
    pev_mac: bytes
    # 6 bytes
    evse_mac: bytes
    # 8 bytes
    run_id: bytes
    # 7 bytes
    nid: bytes
    # 16 bytes
    nmk: bytes
    application_type: int = SLAC_APPLICATION_TYPE
    security_type: int = SLAC_SECURITY_TYPE
    # 2 bytes
    mvf_length: int = 0x56
    # 17 bytes
    pev_id: int = 0x00
    # 17 bytes
    evse_id: int = 0x00
    # 8 bytes
    rsvd_1: int = 0x00
    # 1 bytes
    rsvd_2: int = 0x00

    def __bytes__(self, endianess: str = "big"):
        frame = bytearray(
            self.application_type.to_bytes(1, "big")
            + self.security_type.to_bytes(1, "big")
            + self.mvf_length.to_bytes(2, "little")  # MVF is sent in little endian
            + self.pev_id.to_bytes(17, "big")
            + self.pev_mac
            + self.evse_id.to_bytes(17, "big")
            + self.evse_mac
            + self.run_id
            + self.rsvd_1.to_bytes(8, "big")
            + self.nid
            + self.rsvd_2.to_bytes(1, "big")
            + self.nmk
        )
        if endianess == "big":
            return frame
        return frame.reverse()

    def pack_big(self):
        return self.__bytes__()

    def pack_little(self):
        return self.__bytes__("little")

    @classmethod
    def from_bytes(cls, payload: ctypes) -> "MatchCnf":
        return cls(
            application_type=payload[19],
            security_type=payload[20],
            mvf_length=int.from_bytes(payload[21:23], "big"),
            pev_id=int.from_bytes(payload[23:40], "big"),
            pev_mac=payload[40:46],
            evse_id=int.from_bytes(payload[46:63], "big"),
            evse_mac=payload[63:69],
            run_id=payload[69:77],
            rsvd_1=int.from_bytes(payload[77:85], "big"),
            nid=payload[85:92],
            rsvd_2=payload[92],
            nmk=payload[93:109],
        )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pyslac', 'pyslac.examples', 'pyslac.sockets']

package_data = \
{'': ['*']}

install_requires = \
['environs>=9.5.0,<10.0.0']

setup_kwargs = {
    'name': 'pyslac',
    'version': '0.8.1',
    'description': 'SLAC Protocol implementation, defined in ISO15118-3',
    'long_description': None,
    'author': 'André Duarte',
    'author_email': 'andre@switch-ev.com, andre14x@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



import asyncio
import logging
from binascii import hexlify
from dataclasses import dataclass, field
from inspect import isawaitable
from os import urandom
from typing import List, Optional, Union

from pyslac.enums import (
    ATTEN_RESULTS_TIMEOUT,
    CM_ATTEN_CHAR,
    CM_ATTEN_PROFILE,
    CM_MNBC_SOUND,
    CM_SET_KEY,
    CM_SLAC_MATCH,
    CM_SLAC_PARM,
    CM_START_ATTEN_CHAR,
    ETH_TYPE_HPAV,
    EVSE_PLC_MAC,
    HOMEPLUG_MMV,
    MMTYPE_CNF,
    MMTYPE_IND,
    MMTYPE_REQ,
    MMTYPE_RSP,
    SLAC_ATTEN_TIMEOUT,
    SLAC_GROUPS,
    SLAC_LIMIT,
    SLAC_MSOUNDS,
    SLAC_PAUSE,
    SLAC_RESP_TYPE,
    SLAC_SETTLE_TIME,
    STATE_MATCHED,
    STATE_MATCHING,
    STATE_UNMATCHED,
    FramesSizes,
    Timers,
)

# This timeout is imported from the environment file, because it makes it
# easier to use it with the dev compose file for dev and debugging reasons
from pyslac.environment import Config
from pyslac.layer_2_headers import EthernetHeader, HomePlugHeader
from pyslac.messages import (
    AtennChar,
    AtennCharRsp,
    AttenProfile,
    MatchCnf,
    MatchReq,
    MnbcSound,
    SetKeyCnf,
    SetKeyReq,
    SlacParmCnf,
    SlacParmReq,
    StartAtennChar,
)
from pyslac.sockets.async_linux_socket import (
    create_socket,
    readeth,
    send_recv_eth,
    sendeth,
)
from pyslac.utils import cancel_task, generate_nid, get_if_hwaddr
from pyslac.utils import half_round as hw
from pyslac.utils import task_callback, time_now_ms

logging.basicConfig(level=logging.DEBUG)
logger = logging.getLogger("slac_session")


@dataclass
class SlacSession:
    # pylint: disable=too-many-instance-attributes
    # State can be from EV or EVSE if we use a common data structure,
    # so its init state must come from the constructor
    state: int

    # 16 bytes
    # Network Mask is a random 16 bytes number
    nmk: bytes = b""

    # 7 bytes NetworkIdentifier
    # The 54 LSBs of this field contain the NID (refer to Section 4.4.3.1).
    # The two MSBs shall be set to 0b00.
    # NID is derived from the Network Mask
    nid: bytes = b""

    # FORWARDING_STA
    # 6 bytes (ETHER_ADDR_LEN) IPV6 ADDRESS for the target to send the
    # SLAC sounds responses to
    # associated with ethernet.OSA

    # EVSE-HLE should copy the message OSA to the session variable PEV
    # MAC address; the PEV MAC address can then be used to respond in
    # unicast to the right PEV-HLE; The PEV-HLE address shall also be
    # included in the FORWARDING_STA field of the SLAC MMEs when
    # addressing other GP (GreenPHY) Station (RESP_TYPE = 1);
    # evse_cm_slac_param.c

    # FORWARDING_STA will be 00:00:00:00:00:00 for SLAC when RESP_TYPE=0;
    # the forwarding stations is a vague concept; the specification
    # authors say it should be FF:FF:FF:FF:FF:FF but here, the EVSE-HLE
    # will set it to the PEV MAC (line 110 of evse_cm_slac_param.c)

    # In evse_cm_slac_param.c docstrings it is also said:
    # The PEVHLE address shall also be
    # included in the FORWARDING_STA field of the SLAC MMEs when
    # addressing other GP STA (RESP_TYPE = 1);

    # According to the spec 15118-3, FORWARDING_STA is the EV Host MAC address
    forwarding_sta: bytes = b""

    # PEV_ID (provided by the EV side and received by EVSE on
    # evse_cm_slac_match)
    # 17 bytes
    pev_id: Optional[int] = None

    # 6 bytes
    # EV MAC address, received by EVSE during evse_cm_slac_param, by getting
    # request->ethernet.OSA
    pev_mac: bytes = b""

    # 6 byte EVSE MAC address
    # get channel own MAC in static signed identifier from channel->host
    # received by EV in pev_cm_atten_char for the first time
    evse_mac: bytes = b""

    # 8 bytes identifier used to identify a running session
    # (it is generated by the ev side)
    run_id: bytes = b""
    # 1 byte APPLICATION_TYPE will be 0 for SLAC
    application_type: int = 0x00
    # 1 byte Security Type is also 0x00 for SLAC
    security_type: int = 0x00

    # Counter for the number of CM_START_ATTEN_CHAR.IND received
    # According to ISO15118-3, the EV shall send 3 consecutively
    num_start_attn_rcvd: int = 0

    # Number of total sounds expected to arrive from EV
    # In 15118-3 is associated with CM_EV_match_MNBC
    num_expected_sounds: Optional[int] = None

    # NUM_SOUNDS CM_MNBC_SOUND.IND sent by the EV during SLAC attn charac
    # For the EVSE this variable is incremented during the CM_MNBC_SOUND step
    num_total_sounds: int = 0

    # It is used by the ev (in pev_cm_mnbc_sound.c) to send a determined
    # number of sounds demanded by the evse (SLAC_MSOUNDS).
    sounds: int = SLAC_MSOUNDS

    # Timeout for the reception of Slac sounds
    # The value is 600 ms, but the spec 15118-3 page 38 defines
    # that the value to be transmitted
    # is 0x06, so we divide 600 / 100.
    time_out_ms: int = SLAC_ATTEN_TIMEOUT

    # SLAC_GROUPS = 58 bytes
    # Associated with CM_ATTEN_PROFILE.IND.AAG values defined
    # in evse_cm_mnbc_sound.c
    # if aag: [bytes] = [b'\x00'] * 58
    # an operation with bytes would have to be done like this
    # ag[0] = (int.from_bytes(ag[0], 'big') + 255).to_bytes(1, 'big')
    # So, I guess it is better if it is defined as int from the start and
    # convert to bytes later
    aag: [int] = field(default_factory=lambda: [0] * SLAC_GROUPS)

    # 1byte
    # Number of Slac Groups
    num_groups: Optional[int] = None

    # 17 bytes
    rnd: bytes = (0).to_bytes(17, "big")

    # AttenuationThreshold
    # Limit set in EV side, for checking if it is a match or not
    # 2 or 4 bytes? depends on the architecture
    slac_threshold: int = SLAC_LIMIT

    # This value is used when the ev is sending sound signals to the EVSE.
    # A brief delay (TP_EV_batch_msg_interval) of a few milliseconds is needed
    # between msounds so that EVSE - PLC has time to forward CM_MNBC_SOUND.IND
    # and CM_ATTEN_PROFILE.IND to EVSE-HLE
    pause: int = SLAC_PAUSE

    # Time used by the EVSE and EV to wait after sending a CM_SET_KEY.REQ
    settle_time: int = SLAC_SETTLE_TIME

    # contains the reference to the task running the matching session
    matching_process_task: Optional[asyncio.Task] = None

    def reset(self):
        """
        It resets the session values to their default values
        NID and NMK are not reset, because who handles the operation is the
        call to evse_set_key: as defined by the standard, if we cant set a new
        NID and NMK, then we shall use the already defined ones
        """
        self.state = STATE_UNMATCHED
        self.forwarding_sta = b""
        self.pev_id = None
        self.pev_mac = b""
        self.evse_mac = b""
        self.run_id = b""
        self.application_type = 0x00
        self.security_type = 0x00
        self.num_start_attn_rcvd = 0
        self.num_expected_sounds = None
        self.num_total_sounds = 0
        self.sounds = SLAC_MSOUNDS
        self.time_out_ms = SLAC_ATTEN_TIMEOUT
        self.aag = field(default_factory=lambda: [0] * SLAC_GROUPS)
        self.num_groups = None
        self.rnd = (0).to_bytes(17, "big")
        self.slac_threshold = SLAC_LIMIT
        self.pause = SLAC_PAUSE
        self.settle_time = SLAC_SETTLE_TIME
        self.matching_process_task = None


class SlacEvseSession(SlacSession):
    # pylint: disable=too-many-instance-attributes, too-many-arguments
    # pylint: disable=logging-fstring-interpolation, broad-except
    def __init__(self, evse_id: str, iface: str, config: Config):
        self.iface = iface
        self.evse_id = evse_id
        self.config = config
        host_mac = get_if_hwaddr(self.iface)
        logger.debug(
            f"Session created for evse_id {self.evse_id} on " f"interface {self.iface}"
        )
        self.socket = create_socket(iface=self.iface, port=0)
        self.evse_plc_mac = EVSE_PLC_MAC
        SlacSession.__init__(self, state=STATE_UNMATCHED, evse_mac=host_mac)

    def reset_socket(self):
        self.socket.close()
        self.socket = create_socket(iface=self.iface, port=0)

    async def send_frame(self, frame_to_send: bytes) -> None:
        """
        Async wrapper for a sendeth that checks if sendeth is an awaitable
        """
        # TODO: Add this to a send method
        bytes_sent = sendeth(
            s=self.socket, frame_to_send=frame_to_send, iface=self.iface
        )
        if isawaitable(bytes_sent):
            await bytes_sent

    async def rcv_frame(self, rcv_frame_size: int, timeout: Union[float, int]) -> bytes:
        """
        Helper function to diminush the lines of code when calling the
        asyncio.wait_for with readeth

        :param rcv_frame_size: size of the frame to be received
        :param timeout: timeout for the specific message that is being expected
        :return:
        """
        return await asyncio.wait_for(
            readeth(self.socket, self.iface, rcv_frame_size),
            timeout,
        )

    async def leave_logical_network(self):
        """
        As defined by ISO15118-3 section 9.6, requirement [V2G3-M09-17],
        when leaving the logical network, the parameters associated with
        the current session must be reset to their default values and the
        state of the session shall now be "Unmatched"
        As preparation for the next session, we set a new NMK and NID
        """
        await self.evse_set_key()
        self.reset()

    async def evse_set_key(self) -> bytes:
        """
        PEV-HLE sets the NMK and NID on PEV-PLC using CM_SET_KEY.REQ;
        the NMK and NID must match those provided by EVSE-HLE using
        CM_SLAC_MATCH.CNF;

        The configuration of the low-layer communication module with the
        parameters of the logical network may be done with the MMEs
        CM_SET_KEY.REQ and CM_SET_KEY.CNF.
        Table A.8 from ISO15118-3 defines all the parameters needed and their
        value for this call

        My Nonce for that STA may remain constant for a run of a protocol),
        but a new nonce should be generated for each new protocol run.
        This reflects the purpose of nonces to provide a STA with a quantity i
        t believes to be freshly generated (to defeat replay attacks) and to
        use for association of messages within the protocol run.
        Refer to Section 7.10.7.3 for generation of nonces.

        The only secure way to remove a STA from an AVLN is to change the NMK
        """
        logger.debug("CM_SET_KEY: Started...")
        # for each new set_key message sent (or pyslac session),
        # a new pair of NID (Network ID) and NMK (Network Mask) shall be
        # generated
        nmk = urandom(16)
        # the NID shall be derived from the NMK and its 2 MSBs must be 0b00
        nid = generate_nid(nmk)
        logger.debug("New NMK: %s\n", hexlify(nmk))
        logger.debug("New NID: %s\n", hexlify(nid))
        ethernet_header = EthernetHeader(
            dst_mac=self.evse_plc_mac, src_mac=self.evse_mac
        )
        homeplug_header = HomePlugHeader(CM_SET_KEY | MMTYPE_REQ)
        key_req_payload = SetKeyReq(nid=nid, new_key=nmk)

        frame_to_send = (
            ethernet_header.pack_big()
            + homeplug_header.pack_big()
            + key_req_payload.pack_big()
        )

        # TODO: Change this to just open a socket once for every SlacSession
        # and not every time we call send or send_recv_eth
        # Also think about including the send, rcv method as inner methods of
        # SetKeyReq. Maybe even create a class SetKey that handles both the
        # Send and the CNF of the message
        try:
            await self.send_frame(frame_to_send)
            data_rcvd = await self.rcv_frame(
                rcv_frame_size=FramesSizes.CM_SET_KEY_CNF,
                timeout=Timers.SLAC_INIT_TIMEOUT,
            )
        except asyncio.TimeoutError as e:
            raise TimeoutError("SetKey Timeout raised") from e
        try:
            SetKeyCnf.from_bytes(data_rcvd)
            self.nmk = nmk
            self.nid = nid
        except ValueError as e:
            logger.error(e)
            if self.nmk and self.nid:
                logger.debug(
                    "SetKeyReq has failed, old NMK: %s and NID: %s apply",
                    self.nmk,
                    self.nid,
                )
            else:
                raise ValueError("SetKeyCnf data parsing into the class failed") from e
        logger.debug("Registering NMK and NID into the PLC node...")
        await asyncio.sleep(SLAC_SETTLE_TIME)
        logger.debug("CM_SET_KEY: Finished!")
        return data_rcvd

    async def evse_slac_parm(self) -> None:
        logger.debug("CM_SLAC_PARM: Started...")
        # TODO: Pass the expected parameters later to the read function
        # so that it can be evaluated while the timeout hasnt elapsed
        self.reset_socket()
        while True:
            try:
                # A complete CM_SLAC_PARM.REQ frame must have 60 Bytes:
                # EthernetHeader = 14 bytes
                # HomePlugHeader  = 5 bytes
                # SlacParmReq = 10 bytes
                # Padding = 31 bytes (The min ETH frame must have 60 bytes,
                # it this frame requires padding)
                data_rcvd = await self.rcv_frame(
                    rcv_frame_size=FramesSizes.CM_SLAC_PARM_REQ,
                    timeout=self.config.slac_init_timeout,
                )
            except TimeoutError as e:
                logger.warning(f"Timeout waiting for CM_SLAC_PARM.REQ: {e}")
                raise e
            try:
                ether_frame = EthernetHeader.from_bytes(data_rcvd)
                homeplug_frame = HomePlugHeader.from_bytes(data_rcvd)
                if homeplug_frame.mm_type != CM_SLAC_PARM | MMTYPE_REQ:
                    logger.warning("Frame received is not CM_SLAC_PARM.REQ")
                    logger.debug("Continue waiting for CM_SLAC_PARM.REQ...")
                    continue
                slac_parm_req = SlacParmReq.from_bytes(data_rcvd)
            except Exception as e:
                # TODO: PROPER Exception
                logger.exception(e, exc_info=True)
                raise e
            break

        # Saving SLAC_PARM_REQ parameters from EV
        self.application_type = slac_parm_req.application_type
        self.security_type = slac_parm_req.security_type
        self.run_id = slac_parm_req.run_id

        # both fields are filled with the EV MAC Address
        self.pev_mac = ether_frame.src_mac
        self.forwarding_sta = ether_frame.src_mac

        # SLAC_PARM_CNF frame formation
        ether_header = EthernetHeader(dst_mac=self.pev_mac, src_mac=self.evse_mac)
        homeplug_header = HomePlugHeader(CM_SLAC_PARM | MMTYPE_CNF)
        slac_parm_cnf = SlacParmCnf(forwarding_sta=self.pev_mac, run_id=self.run_id)

        frame_to_send = (
            ether_header.pack_big()
            + homeplug_header.pack_big()
            + slac_parm_cnf.pack_big()
        )

        await self.send_frame(frame_to_send)
        logger.debug("Sent SLAC_PARM.CNF")

        # Update SLAC Session State, indicating that is occupied and ready for
        # a match decision process
        self.state = STATE_MATCHING

        logger.debug("CM_SLAC_PARM: Finished!")

    async def cm_start_atten_charac(self):
        logger.debug("CM_START_ATTEN_CHAR: Started...")
        while True:
            try:
                # A complete CM_START_ATTEN_CHAR.IND frame must have 60 Bytes:
                # EthernetHeader = 14 bytes
                # HomePlugHeader  = 5 bytes
                # StartAtennChar = 19 bytes
                # Padding = 22 bytes (The min ETH frame must have 60 bytes,
                # it this frame requires padding)
                data_rcvd = await self.rcv_frame(
                    rcv_frame_size=FramesSizes.CM_START_ATTEN_CHAR_IND,
                    timeout=Timers.SLAC_REQ_TIMEOUT,
                )
                EthernetHeader.from_bytes(data_rcvd)
                homeplug_frame = HomePlugHeader.from_bytes(data_rcvd)
                if homeplug_frame.mm_type != CM_START_ATTEN_CHAR | MMTYPE_IND:
                    logger.warning("Frame received is not CM_START_ATTEN_CHAR.IND")
                    logger.debug("Continue waiting for CM_START_ATTEN_CHAR.IND...")
                    continue
                start_atten_char = StartAtennChar.from_bytes(data_rcvd)
            except Exception as e:
                logger.exception(e, exc_info=True)
                raise e

            if (
                self.application_type != start_atten_char.application_type
                or self.security_type != start_atten_char.security_type
                or self.run_id != start_atten_char.run_id
                or start_atten_char.resp_type != SLAC_RESP_TYPE
            ):
                logger.exception(ValueError("Error in StartAttenChar"))
                raise ValueError("Error in StartAttenChar")
            break

        # As is stated in ISO15118-3, the EV will send 3 consecutive
        # CM_START_ATTEN_CHAR, regardless if the first one was correctly
        # received and processed. However, the PLC just forwards 1 to the
        # application, so we just need to process 1

        # Saving START_ATTEN_CHAR parameters from EV
        self.num_expected_sounds = start_atten_char.num_sounds
        # the value sent by the EV for the timeout has a factor of 1/100
        # Thus, if the value is e.g. 6, the original value is 600 ms (6 * 100)
        # ATTENTION: Given that Alfen Socket Board interaction with the
        # Smart Controller Board adds a big overhead when sending the data
        # frames, SLAC does not receive all the sounds within
        # `start_atten_char.time_out`. However, according to the following:
        # [V2G3-A09-30] - The EV shall start the timeout timer
        # TT_EV_atten_results (max 1200 ms) when sending the first
        # CM_START_ATTEN_CHAR.IND.
        # [V2G3-A09-31] - While the timer TT_EV_atten_results (max 1200 ms) is
        # running, the EV shall process incoming CM_ATTEN_CHAR.IND messages.
        # Which means, we can use a larger timeout (like 900 ms) so that
        # we receive all or mostly all of the sounds.
        # self.time_out_ms = start_atten_char.time_out * 100
        self.time_out_ms = ATTEN_RESULTS_TIMEOUT
        self.forwarding_sta = start_atten_char.forwarding_sta
        logger.debug("CM_START_ATTEN_CHAR: Finished!")

    def process_sound_frame(
        self,
        homeplug_frame: "HomePlugHeader",
        ether_frame: "EthernetHeader",
        data_rcvd: bytes,
        sounds_rcvd: int,
        aag: List[int],
    ) -> int:
        """
        Helper function that checks which kind of frame was received
        and properly updates the number of sounds received during
        the cm_sounds_loop loop

        returns the next frame size expected
        """
        if homeplug_frame.mm_type == CM_MNBC_SOUND | MMTYPE_IND:
            mnbc_sound_ind = MnbcSound.from_bytes(data_rcvd)
            if self.run_id == mnbc_sound_ind.run_id:
                if self.pev_mac != ether_frame.src_mac:
                    # TODO: Raise Proper Exception
                    raise ValueError(
                        f"Unexpected Source MAC Address for sound "
                        f"number {sounds_rcvd}. "
                        f"PEV MAC: {self.pev_mac}; "
                        f"Source MAC: {ether_frame.src_mac}"
                    )
                logger.debug("MNBC Sound received\n")
                logger.debug("Remaining number of sounds: %s", mnbc_sound_ind.cnt)
            else:
                logger.debug(
                    "Frame received is a CM_MNBC_SOUND but "
                    "it has an invalid Running Session ID. "
                    "Session RunID: %s\n Received RunID: %s",
                    self.run_id,
                    mnbc_sound_ind.run_id,
                )
            return FramesSizes.CM_ATTEN_PROFILE_IND

        if homeplug_frame.mm_type == CM_ATTEN_PROFILE | MMTYPE_IND:
            atten_profile_ind = AttenProfile.from_bytes(data_rcvd)
            if self.pev_mac == atten_profile_ind.pev_mac:
                # Summation of all sounds received per group
                for group in range(atten_profile_ind.num_groups):
                    aag[group] += atten_profile_ind.aag[group]
                self.num_groups = atten_profile_ind.num_groups
                self.num_total_sounds += 1
                logger.debug("ATTEN_Profile Sounds received %s", self.num_total_sounds)
                logger.debug(
                    "Num total sounds: %s / Num expected: %s",
                    self.num_total_sounds,
                    self.num_expected_sounds,
                )
            else:
                logger.warning(
                    "PEV MAC %s does not match: %s. Ignoring...",
                    self.pev_mac,
                    atten_profile_ind.pev_mac,
                )
            return FramesSizes.CM_MNBC_SOUND_IND

    async def cm_sounds_loop(self):
        """
        The GP specification recommends that the EVSE-HLE set an overall
        timer once the cm_start_atten_char message is received and use it
        to terminate the msound loop in case some msounds are lost

        During this process, the EV will send a CM_MNBC_SOUND.IND containing
        a payload that corresponds and is defined within the class MnbcSound as:
        |Application Type|Security Type|SenderID|Cnt|RunID|RSVD|Rnd|

        For each CM_MNBC_SOUND.IND, the EVSE PLC node will send to the host
        application a CM_ATTEN_PROFILE.IND whose payload is defined within
        the class AttenProfile:
        |PEV MAC|NumGroups|RSVD|AAG 1| AAG 2| AAG 3...|

        The sounds reception loop is comprised by the following steps:
        1. awaiting for the reception of a packet
        2. Check for incorrect metadata like Application Type, RunID, ...
        3. Check if the packet is a CM_MNBC_SOUND or CM_ATTEN_PROFILE
        4. if it is a CM_MNBC_SOUND


        accept only CM_MNBC_SOUND.IND that match RunID from the earlier
        CM_SLAC_PARAM.REQ and CM_START_ATTRN_CHAT.IND;

        each CM_MNBC_MSOUND.IND is accompanied by a CM_ATTEN_PROFILE.IND
        but sometimes they arrive out of expected order;

        store the running total of CM_ATTEN_PROFILE.IND.AAG values in
        the session variable and compute the average based on actual
        number of sounds before returning;
        """
        logger.debug("CM_MNBC_SOUND: Started...")
        sounds_rcvd: int = 0
        aag: List[int] = [0] * SLAC_GROUPS
        self.aag = [0] * SLAC_GROUPS
        # time stamp of the start of the signal attenuation measurement and calc
        time_start = time_now_ms()
        self.num_total_sounds = 0
        # We receive in an alternated way the messages
        # CM_MNBC_SOUND.IND and CM_ATTEN_PROFILE.IND and the first
        # message is always a CM_MNBC_SOUND.IND
        next_frame_size: int = FramesSizes.CM_MNBC_SOUND_IND
        while True:
            try:
                data_rcvd = await self.rcv_frame(
                    rcv_frame_size=next_frame_size,
                    # The SLAC_REQ_TIMEOUT used seems to not be enough for the
                    # PLC chip to send a sound, so we use 1 sec instead
                    timeout=1,
                )
                ether_frame = EthernetHeader.from_bytes(data_rcvd)
                homeplug_frame = HomePlugHeader.from_bytes(data_rcvd)
            except Exception as e:
                logger.exception(e, exc_info=True)
                raise e
            if (
                ether_frame.ether_type == ETH_TYPE_HPAV
                and homeplug_frame.mmv == HOMEPLUG_MMV
            ):
                if homeplug_frame.mm_type in [
                    CM_ATTEN_PROFILE | MMTYPE_IND,
                    CM_MNBC_SOUND | MMTYPE_IND,
                ]:
                    next_frame_size = self.process_sound_frame(
                        homeplug_frame, ether_frame, data_rcvd, sounds_rcvd, aag
                    )

                # Check for a timeout of a reception of the expected sounds
                time_elapsed = time_now_ms() - time_start
                if (
                    time_elapsed < self.time_out_ms
                    and self.num_total_sounds < self.num_expected_sounds
                ):
                    continue

                # Time specified by the EV for the Characterization has expired
                # or num of total sounds is >= expected sounds thus, the Atten
                # data must be grouped and averaged before the loop is
                # terminated [V2G3-A09-19]
                if self.num_total_sounds > 0:
                    for group in range(SLAC_GROUPS):
                        self.aag[group] = hw(aag[group] / self.num_total_sounds)
                logger.debug("CM_MNBC_SOUND: Finished!")
                return

    async def cm_atten_char(self):
        logger.debug("CM_ATTEN_CHAR Started...")
        ether_header = EthernetHeader(dst_mac=self.pev_mac, src_mac=self.evse_mac)
        homeplug_header = HomePlugHeader(CM_ATTEN_CHAR | MMTYPE_IND)
        atten_charac = AtennChar(
            source_address=self.pev_mac,
            run_id=self.run_id,
            num_sounds=self.num_total_sounds,
            num_groups=self.num_groups,
            aag=self.aag,
        )

        frame_to_send = (
            ether_header.pack_big()
            + homeplug_header.pack_big()
            + atten_charac.pack_big()
        )

        await self.send_frame(frame_to_send)
        while True:
            try:
                # A complete CM_ATTEN_CHAR.RSP frame must have 70 Bytes:
                # EthernetHeader = 14 bytes
                # HomePlugHeader  = 5 bytes
                # AttenCharRsp = 51 bytes
                data_rcvd = await self.rcv_frame(
                    rcv_frame_size=FramesSizes.CM_ATTEN_CHAR_RSP,
                    # The SLAC_RESP_TIMEOUT used seems to not be enough for the
                    # PLC chip to send a sound, so we use 1 sec instead
                    timeout=1,
                )
                logger.debug(f"Payload Received: \n {hexlify(data_rcvd)}")
                ether_frame = EthernetHeader.from_bytes(data_rcvd)
                homeplug_frame = HomePlugHeader.from_bytes(data_rcvd)
                if homeplug_frame.mm_type != CM_ATTEN_CHAR | MMTYPE_RSP:
                    logger.warning("Frame received is not CM_ATTEN_CHAR.RSP")
                    logger.debug("Continue waiting for CM_ATTEN_CHAR.RSP...")
                    continue
                atten_charac_response = AtennCharRsp.from_bytes(data_rcvd)
            except Exception as e:
                logger.exception(e, exc_info=True)
                raise e

            if (
                ether_frame.ether_type != ETH_TYPE_HPAV
                or homeplug_frame.mmv != HOMEPLUG_MMV
                or self.run_id != atten_charac_response.run_id
            ):
                # TODO: add __str__ or __repr__ methods to the classes
                # for a neat printing
                logger.exception(ether_frame)
                logger.exception(homeplug_frame)
                logger.exception(atten_charac_response)
                # TODO: Check if we shall raise an Error or just ignore
                # According with [V2G3-A09-47] from ISO15118-3, it shall just be
                # ignored
                e = ValueError(
                    "AttenChar Resp Failed, ether type or homeplug "
                    "frame are incorrect."
                )
                logger.exception(e)
                raise e
            break

        if atten_charac_response.result != 0:
            e = ValueError("Atten Char Resp Failed: Atten Char Result " "is not 0x00")
            logger.exception(e)
            raise e
        logger.debug("CM_ATTEN_CHAR: Finished!")

    async def cm_slac_match(self):
        logger.debug("CM_SLAC_MATCH: Started...")
        # Await for a CM_SLAC_MATCH.REQ from EV
        while True:
            try:
                # A complete CM_SLAC_MATCH.REQ frame must have 85 Bytes:
                # EthernetHeader = 14 bytes
                # HomePlugHeader  = 5 bytes
                # AttenCharRsp = 66 bytes
                data_rcvd = await self.rcv_frame(
                    rcv_frame_size=FramesSizes.CM_SLAC_MATCH_REQ,
                    timeout=Timers.SLAC_MATCH_TIMEOUT,
                )

                logger.debug(f"Payload Received: \n {hexlify(data_rcvd)}")
                ether_frame = EthernetHeader.from_bytes(data_rcvd)
                homeplug_frame = HomePlugHeader.from_bytes(data_rcvd)
                if homeplug_frame.mm_type != CM_SLAC_MATCH | MMTYPE_REQ:
                    logger.warning("Frame received is not CM_SLAC_MATCH.REQ")
                    logger.debug("Continue waiting for CM_SLAC_MATCH.REQ...")
                    continue
                slac_match_req = MatchReq.from_bytes(data_rcvd)
            except Exception as e:
                logger.exception(e, exc_info=True)
                raise ValueError("SLAC Match Failed") from e

            if (
                ether_frame.ether_type != ETH_TYPE_HPAV
                or homeplug_frame.mmv != HOMEPLUG_MMV
                or slac_match_req.run_id != self.run_id
            ):
                # TODO: add __str__ or __repr__ methods to the classes
                # for a neat printing
                logger.debug(
                    f"ether_type: {ether_frame.ether_type} \n"
                    f"Expected: {ETH_TYPE_HPAV}"
                )
                logger.debug(
                    f"MMV: {homeplug_frame.mmv} \n " f"Expected: {HOMEPLUG_MMV}"
                )
                logger.debug(
                    f"MMType: {homeplug_frame.mm_type} \n "
                    f"Expected: {CM_SLAC_MATCH | MMTYPE_REQ}"
                )
                logger.debug(
                    f"RunId: {slac_match_req.run_id} \n " f"Expected: {self.run_id}"
                )
                # TODO: Check if we shall raise an Error or just ignore
                # according with requirement [V2G3-A09-98] from ISO15118-3
                # it shall be ignored
                raise ValueError("SLAC Match Request Failed")
            break

        self.pev_id = slac_match_req.pev_id
        self.pev_mac = slac_match_req.pev_mac

        # Send Slac Match Confirmation Message
        ether_header = EthernetHeader(dst_mac=self.pev_mac, src_mac=self.evse_mac)
        homeplug_header = HomePlugHeader(CM_SLAC_MATCH | MMTYPE_CNF)
        slac_match_conf = MatchCnf(
            pev_mac=self.pev_mac,
            evse_mac=self.evse_mac,
            run_id=self.run_id,
            nid=self.nid,
            nmk=self.nmk,
        )

        frame_to_send = (
            ether_header.pack_big()
            + homeplug_header.pack_big()
            + slac_match_conf.pack_big()
        )

        await self.send_frame(frame_to_send)
        logger.debug("CM_SLAC_MATCH: Finished!")
        self.state = STATE_MATCHED

    async def is_link_status_active(self) -> bool:
        """
        This is something I checked that Intec does
        They send a HPGP message called LINK_STATUS.REQ to check if the
        the link between the PEV and EVSE is healthy

        The call is done each 0.5s after the send of a CM_SLAC_MATCH.CNF
        In order to not stress out the chip with requests, we do every 2 secs
        """
        logger.debug("Checking Link Status: Started...")
        ethernet_header = EthernetHeader(
            dst_mac=self.evse_plc_mac, src_mac=self.evse_mac
        )
        LINK_STATUS = 0xA0B8
        mmv = b"\x00"
        mm_type = LINK_STATUS | MMTYPE_REQ
        # Link Status Req does not use the fragmentation fields
        # TODO: Add an option in HomeplgugHeader to add this use case
        homeplug_header_no_fragm = mmv + mm_type.to_bytes(2, "little")
        vendor_mme = 0x00B052
        link_status_req_payload = vendor_mme.to_bytes(3, "big")

        frame_to_send = (
            ethernet_header.pack_big()
            + homeplug_header_no_fragm
            + link_status_req_payload
        )

        # A complete LINK_STATUS.CNF frame must have 60 Bytes:
        # EthernetHeader = 14 bytes
        # HomePlugHeaderNoFrag  = 3 bytes
        # LinkStatusRsp = 3 bytes
        # Padding = 40 bytes (The min ETH frame must have 60 bytes,
        # it this frame requires padding)
        payload_rcvd = send_recv_eth(
            frame_to_send=frame_to_send,
            s=self.socket,
            iface=self.iface,
            rcv_frame_size=FramesSizes.LINK_STATUS_CNF,
        )
        if isawaitable(payload_rcvd):
            payload_rcvd = await payload_rcvd

        logger.debug(f"Payload Received {payload_rcvd}")
        try:
            # TODO: Create the Link Status Class and HomePlug to get properly
            # TODO: the info
            mm_type_rcvd = int.from_bytes(payload_rcvd[15:17], "little")
            if mm_type_rcvd != (LINK_STATUS | MMTYPE_CNF):
                raise ValueError("Message received is not LINK_STATUS.CNF")
        except ValueError as e:
            logger.error(e)
            logger.debug("Link Status: Error")
            return False
        logger.debug("Link Status: Active")
        return True

    async def atten_charac_routine(self):
        await self.cm_start_atten_charac()
        await self.cm_sounds_loop()
        await self.cm_atten_char()
        await self.cm_slac_match()


class SlacSessionController:
    def __init__(self):
        pass

    async def notify_matching_ongoing(self, evse_id: str):
        """
        Used to Notify an external service that a Matching process is ongoing
        """
        pass

    async def notify_matching_failed(self, evse_id: str):
        """
        Used to Notify an external service that a Matching process has failed
        """
        pass

    async def enable_hlc_charging(self, evse_id: str):
        """
        Used to interface with an external service that controls the PWM of the
        Control Pilot circuit and then trigger High Level Communication Charging,
        by enabling the PWM and setting the duty cycle to 5%
        """
        pass

    async def process_cp_state(self, slac_session, state: str):
        """
        If it is the case a matching process is not ongoing
        and the CP has transited to state B, C or D, it spawns a new matching task,
        otherwise if transited to A, E or F and a matching task is running and
        the state is "Matched", then it kills the task. This extra check for the
        state "Matched" is to avoid to kill the task during transitions to state
        E/F which can happen, e.g., if user does EIM after Plugin and before
        the first SLAC message is received.
        """
        # Some states contain the indication if the station can provide energy
        # or not (e.g. A1 - it cant, A2 it can); so we get only the first character
        # from the string, since is that what we are interested here.
        cp_state = state[0]
        logger.debug(f"CP State Received: {state}")
        if cp_state in ["A", "E", "F"] and slac_session.matching_process_task:
            if cp_state == "A" or slac_session.state == STATE_MATCHED:
                # We kill the task if a direct transition to state A is detected
                # or if E,F is detected and we are in state 'Matched'
                await cancel_task(slac_session.matching_process_task)
                logger.debug("Matching process task canceled")
                # leaving the logical network
                # In order to avoid writing too many times to the device,
                # we dont reset the NID and NMK between charging sessions for now
                # await slac_session.leave_logical_network()
                slac_session.matching_process_task = None
                logger.debug("Leaving Logical Network")
        elif cp_state in ["B", "C", "D"] and slac_session.matching_process_task is None:
            slac_session.matching_process_task = asyncio.create_task(
                self.start_matching(slac_session)
            )
            slac_session.matching_process_task.set_name(
                f"Session for EVSE {slac_session.evse_id}"
            )
            # This avoids the exceptions to be "swallowed" by the create_task in the
            # background.
            # TODO: Evaluate the benefits of using frameworks like trio whose event loop
            # forces each task to have a nursery, so that exceptions are not lost
            slac_session.matching_process_task.add_done_callback(task_callback)

    async def start_matching(
        self, slac_session: "SlacEvseSession", number_of_retries=3
    ) -> None:
        """
        Task that is spawned once a state change is detected from A, E or F to
        B, C or D. This task is responsible to run the right methods defined in
        session.py, which as a whole comprise the SLAC protocol.
        In case SLAC fails, it retries up to 3 times to get a match, and
        in case it fails again, it gives up and just with a transition to B, C or D,
        SLAC will restart.

        :param slac_session: Instance of SlacEvseSession
        :param number_of_retries: number of trials before SLAC Mathing is defined
        as a failure
        :return: None
        """
        while number_of_retries:
            number_of_retries -= 1
            await slac_session.evse_slac_parm()
            if slac_session.state == STATE_MATCHING:
                logger.debug("Matching ongoing...")
                await self.notify_matching_ongoing(slac_session.evse_id)
                try:
                    await slac_session.atten_charac_routine()
                except Exception as e:
                    slac_session.state = STATE_UNMATCHED
                    logger.debug(
                        f"Exception Occurred during Attenuation Charc Routine:"
                        f"{e} \n"
                        f"Number of retries left {number_of_retries}"
                    )
            if slac_session.state == STATE_MATCHED:
                logger.debug("PEV-EVSE MATCHED Successfully, Link Established")
                while True:
                    await asyncio.sleep(2.0)

                # The check of the link status wont be done using the message
                # LINK_STATUS, because it is not a proper way to check it since
                # it only provides a confirmation to the message
                # Instead, NW_INFO message is a better one

                # logger.debug("PEV-EVSE Link Lost")
                # leaving the logical network
                # In order to avoid writing too many times to the device,
                # we dont reset the NID and NMK between charging sessions for now
                # await slac_session.leave_logical_network()
                # slac_session.matching_process_task = None
                # break
            if slac_session.state == STATE_UNMATCHED:
                number_of_retries -= 1
                if number_of_retries > 0:
                    logger.warning("PEV-EVSE MATCHED Failed; Retrying..")
                else:
                    logger.error("PEV-EVSE MATCHED Failed: No more retries " "possible")
                    await self.notify_matching_failed(slac_session.evse_id)
            else:
                logger.error(f"SLAC State not recognized {slac_session.state}")

        logger.debug("SLAC Protocol Concluded...")
        # TODO: May need to communicate to HLE that the link is lost (check section
        # 7.5 Loss of communication in -3). Send Unmatched
        # TODO: May need to communicate to CS that the link is gone, so that
        # Basic Charging can be tried
        await slac_session.leave_logical_network()



from dataclasses import dataclass

from pyslac.enums import ETH_TYPE_HPAV, HOMEPLUG_FMID, HOMEPLUG_FMSN, HOMEPLUG_MMV


@dataclass
class EthernetHeader:
    #  6 bytes (channel peer)
    dst_mac: bytes
    #  6 bytes (channel host)
    src_mac: bytes
    #  2 bytes (channel type)
    ether_type: int = ETH_TYPE_HPAV

    def __bytes__(self, endianess: str = "big"):
        if endianess == "big":
            return self.dst_mac + self.src_mac + self.ether_type.to_bytes(2, "big")
        return (
            self.ether_type.to_bytes(2, "little")
            + (int.from_bytes(self.src_mac, "big")).to_bytes(6, "little")
            + (int.from_bytes(self.dst_mac, "big")).to_bytes(6, "little")
        )

    def pack_big(self):
        return self.__bytes__()

    def pack_little(self):
        return self.__bytes__("little")

    @classmethod
    def from_bytes(cls, payload: bytes):
        return cls(
            dst_mac=payload[:6],
            src_mac=payload[6:12],
            ether_type=int.from_bytes(payload[12:14], "big"),
        )


@dataclass
class HomePlugHeader:
    """
    All messages defined in HomePlug GREEN PHY specification Version 1.0 shall
    have the MMV field set to 0x01. HPGP spec page 494

    This payload is defined as follows:
    | MMV | MMTYPE | FMSN | FMID |

    MMV [1 byte] = 0x01
    MMTYPE [2 bytes]: Or operation between the MMID (called MM Base value)
                      to be sent and the mm_type (REQ or CNF)
                      The Management Message Ids can be found in table 11-5
                      page 501 of HPGP standard
    FMSN [1 byte] - Fragmentation Message Sequence number = 0x00
    FMID [1 byte] = 0x00
    """

    mm_type: int
    mmv: bytes = HOMEPLUG_MMV
    fmsn: bytes = HOMEPLUG_FMSN
    fmid: bytes = HOMEPLUG_FMID

    def __bytes__(self, endianess: str = "big"):
        if endianess == "big":
            # The MMType is sent in little endian format
            return self.mmv + self.mm_type.to_bytes(2, "little") + self.fmsn + self.fmid
        return self.fmid + self.fmsn + self.mm_type.to_bytes(2, "little") + self.mmv

    def pack_big(self):
        return self.__bytes__()

    def pack_little(self):
        return self.__bytes__("little")

    @classmethod
    def from_bytes(cls, payload: bytes):
        return cls(
            mmv=payload[14].to_bytes(1, "big"),
            mm_type=int.from_bytes(payload[15:17], "little"),
            fmsn=payload[17].to_bytes(1, "big"),
            fmid=payload[18].to_bytes(1, "big"),
        )



import asyncio
import logging
from ctypes import addressof, create_string_buffer

# pylint: disable=no-name-in-module
from socket import (
    AF_PACKET,
    SO_BROADCAST,
    SOCK_RAW,
    SOL_SOCKET,
    gethostbyname,
    gethostname,
    htons,
    socket,
)
from struct import pack
from typing import Optional

from pyslac.enums import BUFF_MAX_SIZE, Timers
from pyslac.sockets.enums import (
    BPF_ABS,
    BPF_H,
    BPF_JEQ,
    BPF_JMP,
    BPF_K,
    BPF_LD,
    BPF_RET,
    ETH_P_ALL,
    ETH_P_HPAV,
    SO_ATTACH_FILTER,
)
from pyslac.utils import time_now_ms

logging.basicConfig(level=logging.DEBUG)
logger = logging.getLogger("async_linux_socket")


def bpf_jump(code, k, jt, jf):
    return pack("HBBI", code, jt, jf, k)


def bpf_stmt(code, k):
    return bpf_jump(code, k, 0, 0)


# Ordering of the filters is backwards of what would be intuitive for
# performance reasons: the check that is most likely to fail is first.
# A BPF filter code works a bit as assembly code where we have instructions and
# we can jump over them and can load memory positions. The anatomy of a jump
# instruction is as follows:
# bpf_jump(BPF_JMP | BPF_JEQ | BPF_K, <val>, <jtrue>, <jfalse>)
# which basically says, if the value in register BPF_K is equal to <val>,
# then jump <jtrue> instructuins, otherwise jump <jfalse> instructions.
# bpf_stmt(BPF_LD | BPF_H | BPF_ABS, <mem position/ byte offset>)
# this instruction loads to the memory the data present in the register defined
# by the <offset>
# The instruction says Load (BPF_LD) a half word value (BPF_H) in
# from a absolute byte offset (BPF_ABS), in our case, below, that offset is 12.
# An example can be found in
# http://allanrbo.blogspot.com/2011/12/raw-sockets-with-bpf-in-python.html
filters_list = [
    # In this case we want to filter out all ethernet frames whose Ether type
    # is NOT ETH_P_HPAV. Since MAC Dest and Mac source each have 6 bytes,
    # our 2-Octet Ether type is found starting in mem position 12
    # Must be HPGP (check ethertype field at byte offset 12)
    bpf_stmt(BPF_LD | BPF_H | BPF_ABS, 12),
    # If the Ether Type is ETH_P_HPAV, we dont jump and the instruction
    # bpf_stmt(BPF_RET | BPF_K, 0x0fffffff) is run and returns success (-1)
    # otherwise, we jump and we run bpf_stmt(BPF_RET | BPF_K, 0)
    # returning failure (0)
    bpf_jump(BPF_JMP | BPF_JEQ | BPF_K, ETH_P_HPAV, 0, 1),
    bpf_stmt(BPF_RET | BPF_K, 0x0FFFFFFF),  # pass returns -1 in bytes format
    bpf_stmt(BPF_RET | BPF_K, 0),  # reject
]


# TODO:
# Create the socket outside and inject it here
# maybe even create a receive routine inside a task that then sends the
# the frames received using a channel or something


def create_socket(iface: str, port=0) -> socket:
    """
    Creates and binds the raw socket to the desired interface combined with a
    BPF filter

    """
    # Create filters struct and fprog struct to be used by SO_ATTACH_FILTER, as
    # defined in linux/filter.h.
    # An example can be found in
    # http://allanrbo.blogspot.com/2011/12/raw-sockets-with-bpf-in-python.html
    filters = b"".join(filters_list)
    b = create_string_buffer(filters)
    mem_addr_of_filters = addressof(b)
    fprog = pack("HL", len(filters_list), mem_addr_of_filters)

    # https://github.com/spotify/linux/blob/master/include/linux/if_ether.h
    # The link above defines the different protocol (proto) packets that the
    # socket shall receive. By default is 0 (socket.IPPROTO_IP).
    # But we want to receive something, so we use ETH_P_ALL = 0x0003
    # which means receive every packet! (we may want to specify it better
    # the kind)

    # Defining the socket Protocol as ETH_P_ALL, forces the socket to
    # accept all packages during reception (readeth)
    s = socket(AF_PACKET, SOCK_RAW, htons(ETH_P_ALL))
    s.setsockopt(SOL_SOCKET, SO_ATTACH_FILTER, fprog)
    # This option ususally sets up the socket to accept Broadcast messages
    # but I tested without and also works. Nevertheless, let's keep it...
    s.setsockopt(SOL_SOCKET, SO_BROADCAST, 1)
    # The documentation specifies that for the use of the loop socket
    # API, the socket must be non blocking
    # https://docs.python.org/3/library/asyncio-eventloop.html#asyncio.loop.sock_recv
    s.setblocking(False)

    # From the docs: "For raw packet
    # sockets the address is a tuple (ifname, proto [,pkttype [,hatype]])"
    s.bind((iface, port))

    return s


async def sendeth(
    frame_to_send: bytes, iface: Optional[str] = None, port: int = 0, s: socket = None
):
    """Send raw Ethernet packet on interface."""
    loop = asyncio.get_event_loop()

    if not iface:
        iface = gethostbyname(gethostname())

    if not s or not isinstance(s, socket):
        s = create_socket(iface, port)

    padding_bytes = b"\x00" * (60 - len(frame_to_send))
    frame_to_send = frame_to_send + padding_bytes

    return await loop.sock_sendall(s, frame_to_send)


async def readeth_into(s: socket, _):
    """
    https://docs.python.org/3.8/library/asyncio-eventloop.html#asyncio.loop.sock_recv_into

    This alternative requires more testing
    """
    loop = asyncio.get_event_loop()
    data = bytearray(100)
    bytes_rcvd = await loop.sock_recv_into(s, data)
    return data, bytes_rcvd


async def readeth(
    s: socket = None,
    iface: str = None,
    port: int = 0,
    rcv_frame_size: int = BUFF_MAX_SIZE,
    time_start: int = time_now_ms(),
) -> bytes:

    loop = asyncio.get_event_loop()

    if not iface:
        iface = gethostbyname(gethostname())

    if not s or not isinstance(s, socket):
        s = create_socket(iface, port)

    # Maybe I will have to check if the src MAC corresponds to the dst MAC from
    # the sending packet
    bytes_rcvd = await loop.sock_recv(s, rcv_frame_size)
    # NOTE: awaiting for the exact number of bytes expected will just be done,
    # if the rcv frame size is a number below the max ETH PDU possible
    bytes_left = rcv_frame_size - len(bytes_rcvd)
    if bytes_left > 0 and rcv_frame_size < BUFF_MAX_SIZE:
        time_elapsed = time_now_ms() - time_start
        if time_elapsed > Timers.SLAC_INIT_TIMEOUT * 1000:  # in ms
            raise asyncio.TimeoutError
        bytes_rcvd = bytes_rcvd + await readeth(s=s, iface=iface, time_start=time_start)
    return bytes_rcvd


async def send_recv_eth(
    frame_to_send: bytes,
    s: socket = None,
    iface: str = None,
    rcv_frame_size: int = BUFF_MAX_SIZE,
):
    # pylint: disable=lost-exception
    data_rcvd = None
    port = 0

    if not iface:
        iface = gethostbyname(gethostname())

    if not s or not isinstance(s, socket):
        s = create_socket(iface, port)

    try:
        padding_bytes = b"\x00" * (60 - len(frame_to_send))
        frame_to_send = frame_to_send + padding_bytes
        await sendeth(frame_to_send, s=s)
        data_rcvd = await asyncio.wait_for(
            readeth(s=s, iface=iface, rcv_frame_size=rcv_frame_size),
            timeout=Timers.SLAC_INIT_TIMEOUT,
        )
    except asyncio.TimeoutError as e:
        logger.exception(e, exc_info=True)
        s.close()
        logger.info("Linux Socket closed")
        raise e
    finally:
        return data_rcvd
